import warnings
from typing import Union

import geopandas as gpd
import numpy as np

__all__ = ["get_interface_zones", "get_objs"]


def get_objs(
    area: gpd.GeoDataFrame,
    objects: gpd.GeoDataFrame,
    silence_warnings: bool = True,
    predicate: str = "within",
) -> Union[gpd.GeoDataFrame, None]:
    """Get the objects within the area.

    Parameters:
        area (gpd.GeoDataFrame):
            The area of interest in GeoDataFrame.
        objects (gpd.GeoDataFrame):
            The objects (cells) of interest.
        silence_warnings (bool):
            Flag, whether to suppress warnings.
        predicate (str):
            The predicate to use for the spatial join. See `geopandas.tools.sjoin`

    Returns:
        gpd.GeoDataFrame:
            The objects (cells) within the area gdf.
    """
    # NOTE, gdfs need to have same crs, otherwise warning flood.
    inds = objects.geometry.sindex.query(area.iloc[0].geometry, predicate=predicate)
    objs_within: gpd.GeoDataFrame = objects.iloc[np.unique(inds)]

    if objs_within.empty and not silence_warnings:
        warnings.warn(
            """`get_objs_within` resulted in an empty GeoDataFrame. No objects were
            found within the area. Returning None from `get_objs_within`.,
            """,
            RuntimeWarning,
        )
        return

    return objs_within


def get_interface_zones(
    buffer_area: gpd.GeoDataFrame, areas: gpd.GeoDataFrame, buffer_dist: int = 200
) -> gpd.GeoDataFrame:
    """Get the interfaces b/w the polygons defined in a `areas `gdf and `buffer_area`.

    Interface is the region around the border of two touching polygons. The interface
    radius is determined by the `buffer_dist` parameter.

    Applies a buffer to the `buffer_area` and finds the intersection between the buffer
    and the polygons in `areas` gdf.

    Useful for example, when you want to extract the interface of two distinct tissues
    like stroma and cancer.

    Parameters:
        buffer_area (gpd.GeoDataFrame):
            The area or region of interest that is buffered on top of polygons in gdf.
        areas (gpd.GeoDataFrame):
            A geodataframe containing polygons (tissue areas) that might intersect
            with the `buffer_area`.
        buffer_dist (int):
            The radius of the buffer.

    Returns:
        gpd.GeoDataFrame:
            A geodataframe containing the intersecting polygons including the buffer.
    """
    buffer_zone = gpd.GeoDataFrame(
        {"geometry": list(buffer_area.buffer(buffer_dist))},
        crs=buffer_area.crs,
    )
    inter = areas.overlay(buffer_zone, how="intersection")

    # if the intersecting area is covered totally by any polygon in the `areas` gdf
    # take the difference of the intresecting area and the orig roi to discard
    # the roi from the interface 'sheet'
    if not inter.empty:
        if areas.covers(inter.geometry.loc[0]).any():  # len(inter) == 1
            inter = inter.overlay(buffer_area, how="difference")

    return inter


# try:
#     import dask_geopandas

#     _has_dask_geopandas = True
# except ImportError:
#     _has_dask_geopandas = False


# try:
#     from spatialpandas import GeoDataFrame
#     from spatialpandas import sjoin as sp_sjoin
#     from spatialpandas.dask import DaskGeoDataFrame

#     _has_spatialpandas = True
# except ImportError:
#     _has_spatialpandas = False

# try:
#     import dask.distributed  # noqa

#     _has_dask = True
# except ImportError:
#     _has_dask = False
# def get_objs_sp(
#     area: GeoDataFrame,
#     objects: GeoDataFrame,
#     silence_warnings: bool = False,
#     predicate: str = "intersects",
# ) -> GeoDataFrame:
#     """Get objects within the given area.

#     Note:
#         This function is a wrapper around the spatialpandas sjoin function.
#     Note:
#         Optionally parallel execution with Dask.

#     Parameters:
#         area (gpd.GeoDataFrame):
#             The area of interest.
#         objects (gpd.GeoDataFrame):
#             The objects to filter.
#         silence_warnings (bool):
#             Whether to silence warnings.
#         predicate (str):
#             The spatial predicate to use for the spatial join. One of "intersects"

#     Returns:
#         spatialpandas.GeoDataFrame:
#             The objects within the given area. NOTE: This is a spatialpandas GeoDataFrame.
#     """
#     if not _has_spatialpandas:
#         raise ImportError(
#             "spatialpandas not installed. Please install spatialpandas to use this "
#             "function. `pip install spatialpandas`"
#         )

#     if isinstance(objects, DaskGeoDataFrame) and not _has_dask:
#         raise ImportError(
#             "Dask not installed. Please install Dask to use this function in parallel. "
#             '`python -m pip install "dask[distributed]"`'
#         )

#     # run spatialpandas sjoin
#     objs_within = sp_sjoin(objects, area, how="inner", op=predicate)

#     # if dask, compute
#     if isinstance(objects, DaskGeoDataFrame):
#         objs_within = objs_within.compute()

#     if len(objs_within.index) == 0 and not silence_warnings:
#         warnings.warn(
#             """`get_objs_sp` resulted in an empty GeoDataFrame. No objects were
#             found within the area. Returning None from `get_objs_sp`.,
#             """,
#             RuntimeWarning,
#         )
#         return

#     return objs_within


# def get_objs_dgp(
#     area: gpd.GeoDataFrame,
#     objects: dask_geopandas.core.GeoDataFrame,
#     silence_warnings: bool = False,
#     predicate: str = "intersects",
# ) -> GeoDataFrame:
#     """Get objects within the given area.

#     Note:
#         This function is a wrapper around the dask_geopandas sjoin function.

#     Parameters:
#         area (gpd.GeoDataFrame):
#             The area of interest.
#         objects (dask_geopandas.core.GeoDataFrame):
#             The objects to filter.
#         silence_warnings (bool):
#             Whether to silence warnings.
#         predicate (str):
#             The spatial predicate to use for the spatial join. One of "intersects"

#     Returns:
#         gpd.GeoDataFrame:
#             The objects within the given area. NOTE: This is a spatialpandas GeoDataFrame.
#     """
#     if not _has_dask_geopandas:
#         raise ImportError(
#             "dask-geopandas not installed. Please install spatialpandas to use this "
#             "function. `pip install dask-geopandas`"
#         )

#     # run spatialpandas sjoin
#     objs_within = dask_geopandas.sjoin(objects, area, how="inner", predicate=predicate)
#     objs_within: gpd.GeoDataFrame = objs_within.compute()

#     if len(objs_within.index) == 0 and not silence_warnings:
#         warnings.warn(
#             """`get_objs_sp` resulted in an empty GeoDataFrame. No objects were
#             found within the area. Returning None from `get_objs_sp`.,
#             """,
#             RuntimeWarning,
#         )
#         return

#     return objs_within



from pathlib import Path
from typing import Tuple, Union

import geopandas as gpd

from ..utils import pre_proc_gdf, read_gdf
from .save_utils import get_xy_coords

__all__ = ["GSONTile"]


class GSONTile:
    def __init__(
        self,
        filename: Union[Path, str],
        tile_size: Tuple[int, int] = (1000, 1000),
        min_size: int = 30,
    ) -> None:
        """Handle geojson annotations of a tile.

        Allows easy extraction of border and non-border annotations.

        Parameters:
            filename (Union[Path, str]):
                Name of the input geojson file.
            tile_size (Tuple[int, int]):
                Height and width of the tile in pixels.
            min_size (int):
                Minimum size of the annotations in pixels.

        Examples:
            Get non-border annotations of the tile.
            >>> from cellseg_gsontools.tile import GSONTile
            >>> tile = GSONTile("path/to/tile.json")
            >>> tile.non_border_annots
        """
        self.filename = Path(filename)
        self.min_size = min_size
        self.xmin, self.ymin = get_xy_coords(filename)
        self.xmax = self.xmin + tile_size[0]
        self.ymax = self.ymin + tile_size[1]

        self.gdf = pre_proc_gdf(read_gdf(self.filename), min_size=min_size)

    def __len__(self):
        """Return the length of the gson obj."""
        return len(self.gdf)

    def __repr__(self):
        """Return the representation of the gson obj."""
        return f"{self.__class__.__name__}({self.filename})"

    @property
    def non_border_annots(self) -> gpd.GeoDataFrame:
        """Get all the annotations/polygons that do not touch any edges of the tile.

        Note: Resets the index of the returned gdf.
        Note: Origin in the top-left corner of the image/tile. In geoformat, the origin
                is in the bottom-left corner.
        """
        not_r = self.gdf["xmax"] != self.xmax
        not_l = self.gdf["xmin"] != self.xmin
        not_b = self.gdf["ymax"] != self.ymax
        not_t = self.gdf["ymin"] != self.ymin
        non_border_annots = self.gdf[not_r & not_l & not_b & not_t].copy()
        non_border_annots = non_border_annots.reset_index(drop=True)
        return non_border_annots

    @property
    def right_border_annots(self) -> gpd.GeoDataFrame:
        """Get all the annotations/polygons that touch the right edge of the tile.

        Note: Resets the index of the returned gdf.
        Note: Origin in the top-left corner of the image/tile. In geoformat, the origin
                is in the bottom-left corner.
        """
        r_border_anns = self.gdf[self.gdf["xmax"] == self.xmax].copy()

        # translate one unit right
        translated_coords = r_border_anns.translate(xoff=1.0)
        r_border_anns["geometry"] = translated_coords
        r_border_anns = r_border_anns.reset_index(drop=True)

        return r_border_anns

    @property
    def left_border_annots(self) -> gpd.GeoDataFrame:
        """Get all the annotations/polygons that touch the left edge of the tile.

        Note: Resets the index of the returned gdf.
        Note: Origin in the top-left corner of the image/tile. In geoformat, the origin
                is in the bottom-left corner.
        """
        l_border_anns = self.gdf[self.gdf["xmin"] == self.xmin].copy()
        l_border_anns = l_border_anns.reset_index(drop=True)

        return l_border_anns

    @property
    def bottom_border_annots(self) -> gpd.GeoDataFrame:
        """Get all the annotations/polygons that touch the bottom edge of the tile.

        Note: Resets the index of the returned gdf.
        Note: Origin in the top-left corner of the image/tile. In geoformat, the origin
                is in the bottom-left corner.
        """
        b_border_anns = self.gdf[self.gdf["ymax"] == self.ymax].copy()

        # translate 1-unit down
        translated_coords = b_border_anns.translate(yoff=1.0)
        b_border_anns["geometry"] = translated_coords
        b_border_anns = b_border_anns.reset_index(drop=True)

        return b_border_anns

    @property
    def bottom_left_border_annots(self) -> gpd.GeoDataFrame:
        """Get all the annotations/polygons that touch the bottom edge of the tile.

        Note: Resets the index of the returned gdf.
        Note: Origin in the top-left corner of the image/tile. In geoformat, the origin
                is in the bottom-left corner.
        """
        b = (self.gdf["ymax"] == self.ymax) & (self.gdf["xmin"] == self.xmin)
        bl_border_anns = self.gdf[b].copy()

        # translate 1-unit down
        translated_coords = bl_border_anns.translate(yoff=1.0)
        bl_border_anns["geometry"] = translated_coords
        bl_border_anns = bl_border_anns.reset_index(drop=True)

        return bl_border_anns

    @property
    def bottom_right_border_annots(self) -> gpd.GeoDataFrame:
        """Get all the annotations/polygons that touch the bottom edge of the tile."""
        b = (self.gdf["ymax"] == self.ymax) & (self.gdf["xmax"] == self.xmax)
        br_border_anns = self.gdf[b].copy()

        # translate 1-unit down and right
        translated_coords = br_border_anns.translate(yoff=1.0, xoff=1.0)
        br_border_anns["geometry"] = translated_coords
        br_border_anns = br_border_anns.reset_index(drop=True)

        return br_border_anns

    @property
    def top_border_annots(self) -> gpd.GeoDataFrame:
        """Get all the annotations/polygons that touch the top edge of the tile.

        Note: Resets the index of the returned gdf.
        Note: Origin in the top-left corner of the image/tile. In geoformat, the origin
                is in the bottom-left corner.
        """
        t_border_anns = self.gdf[self.gdf["ymin"] == self.ymin].copy()
        t_border_anns = t_border_anns.reset_index(drop=True)

        return t_border_anns

    @property
    def top_right_border_annots(self) -> gpd.GeoDataFrame:
        """Get all the annotations/polygons that touch the top edge of the tile.

        Note: Resets the index of the returned gdf.
        Note: Origin in the top-left corner of the image/tile. In geoformat, the origin
                is in the bottom-left corner.
        """
        b = (self.gdf["ymin"] == self.ymin) & (self.gdf["xmax"] == self.xmax)
        tr_border_anns = self.gdf[b].copy()
        translated_coords = tr_border_anns.translate(xoff=1.0)
        tr_border_anns["geometry"] = translated_coords
        tr_border_anns = tr_border_anns.reset_index(drop=True)

        return tr_border_anns

    @property
    def top_left_border_annots(self) -> gpd.GeoDataFrame:
        """Get all the annotations/polygons that touch the top edge of the tile."""
        b = (self.gdf["ymin"] == self.ymin) & (self.gdf["xmin"] == self.xmin)
        tl_border_anns = self.gdf[b].copy()
        tl_border_anns = tl_border_anns.reset_index(drop=True)

        return tl_border_anns



import geopandas as gpd
import numpy as np

try:
    import geomstats.backend as gs
    from geomstats.geometry.discrete_curves import DiscreteCurves, SRVMetric
    from geomstats.geometry.euclidean import Euclidean
    from geomstats.geometry.pre_shape import PreShapeSpace
    from geomstats.learning.frechet_mean import FrechetMean
except ImportError:
    raise ImportError(
        "`geomstats` package required. Install from github: "
        """git clone https://github.com/geomstats/geomstats.git
           cd geomstats
           pip3 install ."""
    )


M_AMBIENT = 2
r2 = Euclidean(dim=2)
curves_r2 = DiscreteCurves(ambient_manifold=r2)
k_sampling_points = 50
PRESHAPE_SPACE = PreShapeSpace(m_ambient=M_AMBIENT, k_landmarks=k_sampling_points)
PRESHAPE_METRIC = PRESHAPE_SPACE.embedding_space.metric


def center_curves(df: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    """Prepare curves by centering them.

    Parameters
    ----------
    df : gpd.geodataframe
        Dataframe with curves to be centered.

    Returns
    -------
    gpd.geodataframe
    """
    curves = []

    for curve in df["geometry"]:
        x, y = curve.exterior.coords.xy
        x = x - np.median(x)
        y = y - np.median(y)

        centered = list(zip(x, y))
        curves.append(centered)

    curves = np.array(curves)

    df["geometry"] = curves

    return df


def interpolate(contours, n: int = k_sampling_points):
    """Interpolate curve to given number of points.

    Parameters
    ----------
    curve : np.array
        Curve to be interpolated.
    n : np.array
        Number of points for wanted curve.

    Returns
    -------
    np.array
    """
    xc = [a[0] for a in contours]
    yc = [a[1] for a in contours]

    # spacing of x and y points.
    dy = np.diff(yc)
    dx = np.diff(xc)

    # distances between consecutive coordinates
    dS = np.sqrt(dx**2 + dy**2)
    dS = np.append(np.zeros(1), dS)  # include starting point

    # Arc length and perimeter
    d = np.cumsum(dS)
    perim = d[-1]

    # divide the perimeter to evenly spaced values
    ds = perim / n
    dSi = np.arange(0, n) * ds
    dSi[-1] = dSi[-1] - 5e-3

    # interpolate the x and y coordinates
    yi = np.interp(dSi, d, yc)
    xi = np.interp(dSi, d, xc)

    return np.array(list(zip(xi, yi)))


def apply_func_to_ds(input_ds, func):
    """Apply the input function func to the input dictionnary input_ds.

    This function goes through the dictionnary structure and applies
    func to every cell in input_ds[treatment][line].

    It stores the result in a dictionnary output_ds that is returned
    to the user.

    Parameters
    ----------
    input_ds : dict
        Input dictionnary, with keys treatment-line.
    func : callable
        Function to be applied to the values of the dictionnary, i.e.
        the cells.

    Returns
    -------
    output_ds : dict
        Output dictionnary, with the same keys as input_ds.
    """
    output_ds = {}
    output_list = []
    for one_cell in input_ds:
        output_list.append(func(one_cell))
    output_ds = gs.array(output_list)
    return output_ds


def exhaustive_align(curve, base_curve):
    """Align given curve with base curve, orientation minimizing geodesic distance.

    Parameters
    ----------
    curve : np.array
        Curve to be aligned.
    base_curve : np.array
        Curve to be aligned to.

    Returns
    -------
    aligned_curve: gs.array
    """
    nb_sampling = len(curve)
    distances = gs.zeros(nb_sampling)
    base_curve = gs.array(base_curve)
    for shift in range(nb_sampling):
        reparametrized = [curve[(i + shift) % nb_sampling] for i in range(nb_sampling)]
        aligned = PRESHAPE_SPACE.align(
            point=gs.array(reparametrized), base_point=base_curve
        )
        distances[shift] = PRESHAPE_METRIC.norm(
            gs.array(aligned) - gs.array(base_curve)
        )
    shift_min = gs.argmin(distances)
    reparametrized_min = [
        curve[(i + shift_min) % nb_sampling] for i in range(nb_sampling)
    ]
    aligned_curve = PRESHAPE_SPACE.align(
        point=gs.array(reparametrized_min), base_point=base_curve
    )
    return aligned_curve


def frechet_mean(df: gpd.GeoDataFrame, metric=curves_r2.srv_metric):
    """Calculate the Frechet mean in a give metric on an array of curves.

    Parameters
    ----------
    df : gpd.geoDataframe
        Input dictionnary, with keys treatment-line.
    metric : callable
        Function to be applied to the values of the dictionnary, i.e.
        the cells.

    Returns
    -------
    np.array
    """
    ds_proj = [interpolate(curve) for curve in center_curves(df)["geometry"]]

    BASE_CURVE = ds_proj[0]
    ds_align = apply_func_to_ds(ds_proj, func=lambda x: exhaustive_align(x, BASE_CURVE))

    mean = FrechetMean(metric=metric, method="default")
    mean.fit(ds_align)

    mean_estimate = mean.estimate_

    mean_estimate_clean = mean_estimate[~gs.isnan(gs.sum(mean_estimate, axis=1)), :]
    mean_estimate_clean = interpolate(mean_estimate_clean, n=k_sampling_points)

    return mean_estimate_clean


def dists_to_mean(curves, global_mean):
    """Calculate every cells geodesic distance to mean shape of cell collection.

    Parameters
    ----------
    curves : np.array
        Collecion of curves, Prepped and aligned.
    global_mean : np.array
       Calculated Frechet mean of curve collection.

    Returns
    -------
    np.array
    """
    return [SRVMetric.dist(curve, global_mean) for curve in curves]



"""Indices adapted from: https://github.com/pysal/inequality.

BSD 3-Clause License

Copyright (c) 2018, pysal-inequality developers
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from functools import partial
from typing import Sequence, Tuple, Union

import geopandas as gpd
import mapclassify
import numpy as np
from libpysal.weights import W

from .apply import gdf_apply
from .neighbors import neighborhood, nhood_counts, nhood_vals
from .utils import is_categorical, set_uid

__all__ = [
    "simpson_index",
    "shannon_index",
    "gini_index",
    "local_diversity",
]

SMALL = np.finfo("float").tiny


def simpson_index(counts: Sequence) -> float:
    """Compute the Simpson diversity index on a count vector.

    Note:
        Simpson diversity index is a quantitative measure that reflects how many
        different types (such as species) there are in a dataset (a community). It
        is a probability measure, when it is low, the greater the probability that
        two randomly selected individuals will be the same species.
        - [A. Wilson, N. Gownaris](https://bio.libretexts.org/Courses/Gettysburg_College/01%3A_Ecology_for_All/22%3A_Biodiversity/22.02%3A_Diversity_Indices)


    **Simpson index:**
    $$
    D = 1 - \\sum_{i=1}^n \\left(\\frac{n_i}{N}\\right)^2
    $$

    where $n_i$ is the count of species $i$ and $N$ is the total count of species.

    Parameters:
        counts (Sequence):
            A count vector/list of shape (C, ).

    Returns:
        float:
            The computed Simpson diversity index.
    """
    N = np.sum(counts) + SMALL
    return 1 - np.sum([(n / N) ** 2 for n in counts if n != 0])


def shannon_index(counts: Sequence) -> float:
    """Compute the Shannon Weiner index/entropy on a count vector.

    Note:
        "*The Shannon index is related to the concept of uncertainty. If for example,
        a community has very low diversity, we can be fairly certain of the identity of
        an organism we might choose by random (high certainty or low uncertainty). If a
        community is highly diverse and we choose an organism by random, we have a
        greater uncertainty of which species we will choose (low certainty or high
        uncertainty).*"
        - [A. Wilson, N. Gownaris](https://bio.libretexts.org/Courses/Gettysburg_College/01%3A_Ecology_for_All/22%3A_Biodiversity/22.02%3A_Diversity_Indices)

    **Shannon index:**
    $$
    H^{\\prime} = -\\sum_{i=1}^n p_i \\ln(p_i)
    $$

    where $p_i$ is the proportion of species $i$ and $n$ is the total count of species.

    Parameters:
        counts (Sequence):
            A count vector/list of shape (C, ).

    Returns:
        float:
            The computed Shannon diversity index.
    """
    N = np.sum(counts) + SMALL
    probs = [float(n) / N for n in counts]

    entropy = -np.sum([p * np.log(p) for p in probs if p != 0])

    if entropy == 0:
        return 0.0

    return entropy


def gini_index(x: Sequence) -> float:
    """Compute the gini coefficient of inequality for species.

    Note:
        This is based on
        http://www.statsdirect.com/help/default.htm#nonparametric_methods/gini.htm.

    **Gini-index:**
    $$
    G = \\frac{\\sum_{i=1}^n (2i - n - 1)x_i} {n \\sum_{i=1}^n x_i}
    $$

    where $x_i$ is the count of species $i$ and $n$ is the total count of species.

    Parameters:
        x (Sequence):
            The input value-vector. Shape (n, )

    Raises:
        ValueError:
            If there are negative input values.

    Returns:
        float:
            The computed Gini coefficient.
    """
    if np.min(x) < 0:
        raise ValueError("Input values need to be positive for Gini coeff")

    n = len(x)
    s = np.sum(x)
    nx = n * s + SMALL

    rx = (2.0 * np.arange(1, n + 1) * x[np.argsort(x)]).sum()
    return (rx - nx - s) / nx


def theil_index(x: Sequence) -> float:
    """Compute the Theil index of inequality for species.

    **Theil-index:**
    $$
    T = \\sum_{i=1}^n \\left(
        \\frac{y_i}{\\sum_{i=1}^n y_i} \\ln \\left[
            N \\frac{y_i} {\\sum_{i=1}^n y_i}
        \\right]\\right)
    $$

    where $y_i$ is the count of species $i$ and $N$ is the total count of species.

    Parameters:
        x (Sequence):
            The input value-vector. Shape (n, )

    Returns:
        float:
            The computed Theil index.
    """
    n = len(x)
    x = x + SMALL * (x == 0)  # can't have 0 values
    xt = np.sum(x, axis=0) + SMALL
    s = x / (xt * 1.0)
    lns = np.log(n * s)
    slns = s * lns
    t = np.sum(slns)

    return t


def theil_between_group(x: Sequence, partition: Sequence) -> float:
    """Compute the between group Theil index.

    Parameters:
        x (Sequence):
            The input value-vector. Shape (n, )
        partition (Sequence):
            The groups for each x value. Shape (n, ).

    Returns:
        float:
            The computed between group Theil index.
    """
    groups = np.unique(partition)
    x_total = x.sum(0) + SMALL

    # group totals
    g_total = np.array([x[partition == gid].sum(axis=0) for gid in groups])

    if x_total.size == 1:  # y is 1-d
        sg = g_total / (x_total * 1.0)
        sg.shape = (sg.size, 1)
    else:
        sg = np.dot(g_total, np.diag(1.0 / x_total))

    ng = np.array([np.sum(partition == gid) for gid in groups])
    ng.shape = (ng.size,)  # ensure ng is 1-d
    n = x.shape[0]

    # between group inequality
    sg = sg + (sg == 0)  # handle case when a partition has 0 for sum
    bg = np.multiply(sg, np.log(np.dot(np.diag(n * 1.0 / ng), sg))).sum(axis=0)

    return float(bg)


def theil_within_group(x: Sequence, partition: Sequence) -> float:
    """Compute the within group Theil index.

    Parameters:
        x (Sequence):
            The input value-vector. Shape (n, )
        partition (Sequence):
            The groups for each x value. Shape (n, ).

    Returns:
        float:
            The computed within group Theil index.
    """
    theil = theil_index(x)
    theil_bg = theil_between_group(x, partition)

    return float(theil - theil_bg)


DIVERSITY_LOOKUP = {
    "simpson_index": simpson_index,
    "shannon_index": shannon_index,
    "gini_index": gini_index,
    "theil_index": theil_index,
}


GROUP_DIVERSITY_LOOKUP = {
    "theil_between_group": theil_between_group,
    "theil_within_group": theil_within_group,
}


def local_diversity(
    gdf: gpd.GeoDataFrame,
    spatial_weights: W,
    val_col: Union[str, Tuple[str, ...]],
    id_col: str = None,
    metrics: Tuple[str, ...] = ("simpson_index",),
    scheme: str = "FisherJenks",
    parallel: bool = True,
    num_processes: int = -1,
    rm_nhood_cols: bool = True,
    col_prefix: str = None,
    create_copy: bool = True,
) -> gpd.GeoDataFrame:
    """Compute the local diversity/heterogenity metric for cell neighborhood.

    Note:
        Allowed diversity metrics:

        - `simpson_index` - for both categorical and real valued neighborhoods
        - `shannon_index` - for both categorical and real valued neighborhoods
        - `gini_index` - for only real valued neighborhoods
        - `theil_index` - for only real valued neighborhoods

    Note:
        If `val_col` is not categorical, the values are binned using `mapclassify`.
        The bins are then used to compute the diversity metrics. If `val_col` is
        categorical, the values are used directly.

    Parameters:
        gdf (gpd.GeoDataFrame):
            The input GeoDataFrame.
        spatial_weights (libysal.weights.W):
            Libpysal spatial weights object.
        val_col (Union[str, Tuple[str, ...]]):
            The name of the column in the gdf for which the diversity is computed.
            You can also pass in a list of columns, in which case the diversity is
            computed for each column.
        id_col (str):
            The unique id column in the gdf. If None, this uses `set_uid` to set it.
            Defaults to None.
        metrics (Tuple[str, ...]):
            A Tuple/List of diversity metrics. Allowed metrics: "shannon_index",
            "simpson_index", "gini_index", "theil_index". Defaults to None.
        scheme (str):
            `mapclassify` classification scheme. Defaults to "FisherJenks".
        parallel (bool):
            Flag whether to use parallel apply operations when computing the diversities.
            Defaults to True.
        num_processes (int, default=-1):
            The number of processes to use when parallel=True. If -1,
            this will use all available cores.
        rm_nhood_cols (bool):
            Flag, whether to remove the extra neighborhood columns from the result gdf.
            Defaults to True.
        col_prefix (str):
            Prefix for the new column names. Defaults to None.
        create_copy (bool):
            Flag whether to create a copy of the input gdf or not. Defaults to True.

    Raises:
        ValueError:
            If an illegal metric is given.

    Returns:
        gpd.GeoDataFrame:
            The input geodataframe with computed diversity metric columns added.

    Examples:
        Compute the simpson diversity of eccentricity values for each cell neighborhood
        >>> from cellseg_gsontools.diversity import local_diversity
        >>> from cellseg_gsontools.graphs import fit_graph
        >>> w = fit_graph(gdf, type="distband", thres=75.0)
        >>> local_diversity(
        ...     gdf,
        ...     spatial_weights=w_dist,
        ...     val_col="eccentricity",
        ...     metrics=["simpson_index"],
        ... )
    """
    allowed = list(DIVERSITY_LOOKUP.keys())
    if not all(m in allowed for m in metrics):
        raise ValueError(
            f"Illegal metric in `metrics`. Got: {metrics}. Allowed metrics: {allowed}."
        )

    if create_copy:
        gdf = gdf.copy()

    # set uid
    if id_col is None:
        id_col = "uid"
        gdf = set_uid(gdf)

    # If shannon or simpson index in metrics, counts are needed
    ret_counts = False
    if any([m in metrics for m in ("simpson_index", "shannon_index")]):
        ret_counts = True

    # If Gini is in metrics, neighboring values are needed
    gt = ("gini_index", "theil_index")
    ret_vals = False
    if any([m in metrics for m in gt]):
        ret_vals = True

    # Get the immediate node neighborhood
    func = partial(neighborhood, spatial_weights=spatial_weights)
    gdf["nhood"] = gdf_apply(
        gdf,
        func,
        columns=[id_col],
        axis=1,
        parallel=parallel,
        num_processes=num_processes,
    )

    if isinstance(val_col, str):
        val_col = (val_col,)

    for col in val_col:
        values = gdf[col]

        # Get bins if data not categorical
        if not is_categorical(values):
            bins = mapclassify.classify(values, scheme=scheme).bins
        else:
            bins = None

        # Get the counts of the binned metric inside the neighborhoods
        if ret_counts:
            func = partial(nhood_counts, values=values, bins=bins)
            gdf[f"{col}_nhood_counts"] = gdf_apply(
                gdf,
                func,
                columns=["nhood"],
                axis=1,
                parallel=parallel,
                num_processes=num_processes,
            )

        if ret_vals:
            func = partial(nhood_vals, values=values)
            gdf[f"{col}_nhood_vals"] = gdf_apply(
                gdf,
                func,
                columns=["nhood"],
                axis=1,
                parallel=parallel,
                num_processes=num_processes,
            )

        # Compute the diversity metrics for the neighborhood counts
        for metric in metrics:
            colname = f"{col}_nhood_counts" if metric not in gt else f"{col}_nhood_vals"

            col_prefix = "" if col_prefix is None else col_prefix
            gdf[f"{col_prefix}{col}_{metric}"] = gdf_apply(
                gdf,
                DIVERSITY_LOOKUP[metric],
                columns=[colname],
                parallel=parallel,
                num_processes=num_processes,
            )

        if rm_nhood_cols:
            gdf = gdf.drop(labels=[colname], axis=1)

    if rm_nhood_cols:
        gdf = gdf.drop(labels=["nhood"], axis=1)

    return gdf



from functools import partial
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

import geopandas as gpd
import networkx as nx
import pandas as pd
from libpysal.weights import W, fuzzy_contiguity
from shapely.ops import unary_union
from tqdm import tqdm

from ..multiproc import run_pool
from ..utils import read_gdf, set_uid
from ._base_merger import BaseGSONMerger
from .poly_utils import validate_and_simplify
from .save_utils import gdf_to_file

__all__ = ["AreaMerger"]


class AreaMerger(BaseGSONMerger):
    """Merge the area/tissue annotation files of the tiles to one file.

    Note:
        Assumes:

        - Input files contain area/tissue semantic segmentation annotations.
        - Files have start x and y coords embedded in filename e.g. `x-[coord]_y-[coord]`
        - Tiles are the same size.
        - Allowed input file-formats `.json`, `.geojson`, `.feather`, `.parquet`

    Parameters:
        in_dir (Union[Path, str]):
            Path to the directory containing the annotation files of tiles.

    Attributes:
        annots (gpd.GeoDataFrame):
            A gdf of the resulting annotations. Available after merging.
    """

    def __init__(self, in_dir: Union[Path, str]) -> None:
        super().__init__(in_dir, None)

    def merge_dir(
        self,
        out_fn: Optional[Union[Path, str]] = None,
        format: Optional[str] = None,
        verbose: bool = True,
        parallel: bool = False,
    ) -> None:
        """Merge all the semantic segmentation files in the input directory into one.

        Note:
            Unlike the CellMerger.merge_dir() -method, this can be parallelized with
            `parallel=True` -argument.

        Parameters:
            out_fn (Union[Path, str]):
                Filename for the output file. If None, the merged gdf is saved to the
                class attribute `self.annots` only.
            format (str):
                The format of the output geojson file. One of: "feather", "parquet",
                "geojson", None. This is ignored if `out_fn` is None.
            verbose (bool):
                Whether to show a progress bar or not.
            parallel (bool):
                Whether to use parallel processing or not.

        Examples:
            Write feather files to a '.geojson' file.
            >>> from cellseg_gsontools.merging import AreaMerger
            >>> merger = AreaMerger("/path/to/feather_files/")
            >>> merger.merge_dir("/path/to/output.json", format="geojson")

            Write input geojson files to feather file.
            >>> from cellseg_gsontools.merging import AreaMerger
            >>> merger = AreaMerger("/path/to/geojsons/")
            >>> merger.merge_dir("/path/to/output.feather", format="feather")

            Write input parquet files to parquet file.
            >>> from cellseg_gsontools.merging import AreaMerger
            >>> merger = AreaMerger("/path/to/parquet_files/")
            >>> merger.merge_dir("/path/to/output.parquet", format="parquet")
        """
        if out_fn is not None:
            out_fn = Path(out_fn)

        if format not in (".feather", ".parquet", ".geojson", None):
            raise ValueError(
                f"Invalid format. Got: {format}. Allowed: .feather, .parquet, .geojson"
            )

        # merge the tiles
        self.annots = self._merge(verbose=verbose, parallel=parallel)

        if verbose:
            msg = f"{format}-format" if out_fn is not None else "`self.annots`"
            print(f"Saving the merged geojson file: {out_fn} to {msg}")

        if out_fn is not None:
            # save the merged geojson
            gdf_to_file(self.annots, out_fn, format)

    def _read_files_to_gdf(
        self, files: List[Path], verbose: bool = True
    ) -> gpd.GeoDataFrame:
        """Read in the input files to a gdf."""
        cols = None
        rows = []
        pbar = tqdm(files, total=len(files)) if verbose else files
        for file in pbar:
            gdf = read_gdf(file)
            if gdf is not None and not gdf.empty:
                cols = gdf.columns
                for _, row in gdf.iterrows():
                    rows.append(row)

        return gpd.GeoDataFrame(rows, columns=cols)

    def _merge_adjascent_polygons(
        self, gdf: gpd.GeoDataFrame, w: W, verbose: bool = True, cl: str = None
    ) -> gpd.GeoDataFrame:
        """Divide spatial weights into subgraphs & merge the polygons in each."""
        # Get all disconnected subgraphs.
        G = w.to_networkx()
        sub_graphs = [
            W(nx.to_dict_of_lists(G.subgraph(c).copy()))
            for c in nx.connected_components(G)
        ]

        # loop over the subgraphs
        pbar = tqdm(sub_graphs, total=len(sub_graphs)) if verbose else sub_graphs
        result_polygons = []
        for sub_w in pbar:
            if verbose:
                pbar.set_description(f"Processing {cl} connected regions:")

            # init a visited lookup table for nodes
            visited = {node: False for node in sub_w.neighbors.keys()}

            # loop over the nodes/polygons in the subgraph
            polygons_to_merge = []
            for node, neighs in sub_w.neighbors.items():
                # if an island, buffer the polygon
                if not neighs:
                    poly = gdf.loc[node].geometry.buffer(2)
                    result_polygons.append(
                        validate_and_simplify(poly, simplify=True, buffer=0.0)
                    )
                    continue

                # if not visited, check if it intersects with any of its neighbors
                # and add it to the list of polygons to merge
                if not visited[node]:
                    poly = gdf.loc[node].geometry.buffer(2)
                    # poly = validate_and_simplify(poly, simplify=True, buffer=0.0)
                    inter = [poly]
                    for neigh in neighs:
                        if not visited[neigh]:
                            neigh_poly = gdf.loc[neigh].geometry.buffer(2)
                            neigh_poly = validate_and_simplify(
                                poly, simplify=True, buffer=0.0
                            )
                            if poly.intersects(neigh_poly):
                                inter.append(neigh_poly)
                                visited[neigh_poly] = True

                    visited[node] = True
                    polygons_to_merge.extend(inter)

            # don't merge if there are no polygons to merge
            if polygons_to_merge:
                result_polygons.append(unary_union(polygons_to_merge))

        # return the merged polygons as gdf
        out = gpd.GeoDataFrame({"geometry": result_polygons, "class_name": cl})
        return out[~out.is_empty].explode(index_parts=False).reset_index(drop=True)

    def _merge_one(
        self, in_gdf: gpd.GeoDataFrame, cl: str, verbose: bool = True
    ) -> gpd.GeoDataFrame:
        """Merge the polygons in one class."""
        in_gdf = set_uid(in_gdf)

        w = fuzzy_contiguity(
            in_gdf,
            buffering=True,
            buffer=2,
            predicate="intersects",
            silence_warnings=True,
        )

        return self._merge_adjascent_polygons(in_gdf, w, verbose=verbose, cl=cl)

    def _merge_one_wrap(self, args: Tuple[Any], verbose: bool = True):
        return self._merge_one(*args, verbose=verbose)

    def _merge(self, verbose: bool = True, parallel: bool = True) -> gpd.GeoDataFrame:
        """Merge the polygons in by class."""
        gdf = self._read_files_to_gdf(self.files)
        classes = gdf["class_name"].unique()

        if not parallel:
            polys = []
            for cl in classes:
                in_gdf = gdf[gdf["class_name"] == cl]
                polys.append(self._merge_one(in_gdf, cl, verbose=verbose))
        else:
            merge_func = partial(self._merge_one_wrap, verbose=False)

            polys = run_pool(
                merge_func,
                [(gdf[gdf["class_name"] == cl], cl) for cl in classes],
                n_jobs=len(classes),
                pbar=verbose,
                pooltype="thread",
                maptype="uimap",
            )

        return set_uid(pd.concat(polys, ignore_index=True))



from typing import Dict, Tuple

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
from libpysal.weights import W

__all__ = ["plot_graph"]


# This is the W.plot() class-methsd from PySAL, only added figsize param.
# https://github.com/pysal/libpysal/blob/main/libpysal/weights/weights.py
def plot_graph(
    gdf: gpd.GeoDataFrame,
    w: W,
    indexed_on: str = None,
    ax: plt.Axes = None,
    node_kws: Dict = None,
    edge_kws: Dict = None,
    figsize: Tuple[float, float] = (10, 10),
) -> Tuple[plt.Figure, plt.Axes]:
    """Plot a graph from a GeoDataFrame and a PySAL W object.

    Parameters
    ----------
    gdf : gpd.GeoDataFrame
        The GeoDataFrame to plot.
    w : W
        The PySAL W object that defines the graph.
    indexed_on : str, optional
        The column in `gdf` to use as the index for the graph, by default None
    ax : plt.Axes, optional
        The axes to plot on, by default None
    node_kws : Dict, optional
        Keyword arguments to pass to the node plotting function, by default None
    edge_kws : Dict, optional
        Keyword arguments to pass to the edge plotting function, by default None
    figsize : Tuple[float, float], optional
        The size of the figure to plot, by default (10, 10).

    Returns
    -------
    Tuple[plt.Figure, plt.Axes]
        The figure and axes used for plotting.

    Examples
    --------
    >>> import geopandas as gpd
    >>> import libpysal
    """
    if ax is None:
        f = plt.figure(figsize=figsize)
        ax = plt.gca()
    else:
        f = plt.gcf()

    color = "k"
    if node_kws is None:
        node_kws = dict(color=color)
    if edge_kws is None:
        edge_kws = dict(color=color)

    for idx, neighbors in w.neighbors.items():
        # skip islands
        if idx in w.islands:
            continue

        if indexed_on is not None:
            neighbors = gdf[gdf[indexed_on].isin(neighbors)].index.tolist()
            idx = gdf[gdf[indexed_on] == idx].index.tolist()[0]

        centroids = gdf.loc[neighbors].centroid.apply(lambda p: (p.x, p.y))
        centroids = np.vstack(centroids.values)
        focal = np.hstack(gdf.loc[idx].geometry.centroid.xy)

        seen = set()
        for nidx, neighbor in zip(neighbors, centroids):
            if (idx, nidx) in seen:
                continue
            ax.plot(*list(zip(focal, neighbor)), marker=None, **edge_kws)
            seen.update((idx, nidx))
            seen.update((nidx, idx))

    ax.scatter(
        gdf.centroid.apply(lambda p: p.x),
        gdf.centroid.apply(lambda p: p.y),
        **node_kws,
    )

    return f, ax



from pathlib import Path

BASE_PATH = Path(__file__).parent.resolve()


def _load(f):
    """Load a gdf file located in the data directory.

    Parameters:
        f (str):
            File name.

    Returns:
        gpd.GeoDataFrame:
            A gdf loaded from f.
    """
    from cellseg_gsontools.utils import read_gdf

    return read_gdf(f)


def cervix_tissue():
    """A GeoDataframe containing cervical tissue areas.

    Examples:
        >>> from cellseg_gsontools.data import cervix_tissue
        >>> cervix_tissue().head(3)
                type                                           geometry  class_name
        uid
        1    Feature  POLYGON ((1852.953 51003.603, 1853.023 51009.1...  areastroma
        2    Feature  POLYGON ((4122.334 48001.899, 4122.994 48014.8...   areagland
        3    Feature  POLYGON ((3075.002 48189.068, 3075.001 48218.8...   areagland
    """
    return _load(BASE_PATH / "cervix_tissue.feather")


def cervix_cells():
    """A GeoDataframe containing cells of the cervical tissue.

    Examples:
        >>> from cellseg_gsontools.data import cervix_cells
        >>> cervix_cells().head(3)
                type                                           geometry    class_name
        uid
        1    Feature  POLYGON ((-10.988 48446.005, -10.988 48453.996...  inflammatory
        2    Feature  POLYGON ((-20.988 48477.996, -19.990 48479.993...    connective
        3    Feature  POLYGON ((-14.988 48767.995, -11.993 48770.990...  inflammatory
    """
    return _load(BASE_PATH / "cervix_cells.feather")


def gland_cells():
    """A GeoDataframe containing cells of cervical gland tissue.

    Examples:
        >>> from cellseg_gsontools.data import gland_cells
        >>> gland_cells().plot()
        plt.Axes
    """
    return _load(BASE_PATH / "gland_cells.feather")


def gland_tissue():
    """A GeoDataframe containing cervical gland tissue.

    Examples:
        >>> from cellseg_gsontools.data import gland_tissue
        >>> gland_tissue().plot()
        plt.Axes
    """
    return _load(BASE_PATH / "gland_areas.feather")


def tumor_stroma_intreface_cells():
    """A GeoDataframe containing cells of a TS-interface.

    Examples:
        >>> from cellseg_gsontools.data import tumor_stroma_intreface_cells
        >>> tumor_stroma_intreface_cells().plot()
        plt.Axes
    """
    return _load(BASE_PATH / "ts_iface_cells.feather")


def tumor_stroma_intreface_tissue():
    """A GeoDataframe containing tissues of a TS-interface.

    Examples:
        >>> from cellseg_gsontools.data import tumor_stroma_intreface_tissue
        >>> tumor_stroma_intreface_tissue().plot()
        plt.Axes
    """
    return _load(BASE_PATH / "ts_iface_areas.feather")


def tissue_merge_dir():
    """Return the path to the directory containing the adjascent tissue patches."""
    return BASE_PATH / "merge_data" / "area"


def cell_merge_dir():
    """Return the path to the directory containing the adjascent cell patches."""
    return BASE_PATH / "merge_data" / "cell"



from typing import List, Tuple

import numpy as np
from skimage.morphology import dilation, disk, erosion

__all__ = ["bounding_box", "crop_to_bbox", "maskout_array"]


def bounding_box(mask: np.ndarray) -> List[int]:
    """Bounding box coordinates for an instance that is given as input.

    This assumes that the `inst_map` has only one instance in it.

    Parameters:
        inst_map (np.ndarray):
            Instance labelled mask. Shape (H, W).

    Returns:
        List[int]:
            List of the origin- and end-point coordinates of the bbox.
    """
    rows = np.any(mask, axis=1)
    cols = np.any(mask, axis=0)
    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]

    rmax += 1
    cmax += 1

    return [rmin, rmax, cmin, cmax]


def crop_to_bbox(
    src: np.ndarray, mask: np.ndarray, dilation_level: int = 0
) -> Tuple[np.ndarray, np.ndarray]:
    """Crops an image and mask to the bounding box of the mask.

    Parameters:
        src (np.ndarray):
            Source image. Shape (H, W, 3).
        mask (np.ndarray):
            Mask to crop the image with. Shape (H, W).
        dilation_level (int):
            Dilation level for the mask.

    Raises:
        ValueError: If the src array is not 2D or 3D.

    Returns:
        Tuple[np.ndarray, np.ndarray]:
            Cropped image and mask.
    """
    if not 2 <= src.ndim <= 3:
        raise ValueError("src must be a 2D or 3D array.")

    if dilation_level > 0:
        mask = dilation(mask, disk(dilation_level))

    ymin, ymax, xmin, xmax = bounding_box(mask)

    # erode back to orig mask
    if dilation_level > 0:
        mask = erosion(mask, disk(dilation_level))

    mask = mask[ymin:ymax, xmin:xmax]
    src = src[ymin:ymax, xmin:xmax]

    return src, mask


def maskout_array(
    src: np.ndarray,
    mask: np.ndarray,
) -> np.ndarray:
    """Masks out the input array with the given mask."""
    if not 2 <= src.ndim <= 3:
        raise ValueError("src must be a 2D or 3D array.")

    if src.ndim == 3:
        src = src * mask[..., None]
    else:
        src = src * mask

    return src



import geopandas as gpd
import numpy as np
import pandas as pd
from sklearn.cluster import DBSCAN, HDBSCAN, OPTICS

__all__ = ["cluster_points"]


def cluster_points(
    gdf: gpd.GeoDataFrame,
    eps: float = 350.0,
    min_samples: int = 30,
    method: str = "dbscan",
    n_jobs: int = -1,
    **kwargs,
) -> gpd.GeoDataFrame:
    """Apply a clustering to centroids in a gdf.

    This is just a quick wrapper for a few clustering algos adapted
    to geodataframes.

    Note:
        Allowed clustering methods are:

        - `dbscan` (sklearn.cluster.DBSCAN)
        - `hdbscan` (sklearn.cluster.HDBSCAN)
        - `optics` (sklearn.cluster.OPTICS)
        - `adbscan` (esda.adbscan.ADBSCAN)

    Parameters:
        gdf (gpd.GeoDataFrame):
            Input geo dataframe with a properly set geometry column.
        eps (float):
            The maximum distance between two samples for one to be considered as in the
            neighborhood of the other. This is not a maximum bound on the distances of
            gdf within a cluster.
        min_samples (int):
            The number of samples (or total weight) in a neighborhood for a point to be
            considered as a core point. This includes the point itself.
        method (str):
            The clustering method to be used. Allowed: ("dbscan", "adbscan", "optics").
        n_jobs (int):
            The number of parallel jobs to run. None means 1. -1 means using all
            processors.
        **kwargs (Dict[str, Any]):
            Arbitrary key-word arguments passed to the clustering methods.

    Raises:
        ValueError:
            If illegal method is given or input `gdf` is of wrong type.

    Returns:
        gpd.GeoDataFrame:
            The input gdf with a new "labels" columns of the clusters.

    Examples:
        Cluster immune cell centroids in a gdf using dbscan.

        >>> from cellseg_gsontools.clustering import cluster_points
        >>> gdf = read_gdf("cells.json")
        >>> gdf = cluster_points(
        ...     gdf[gdf["class_name"] == "immune"],
        ...     method="dbscan",
        ...     eps=350.0,
        ...     min_samples=30,
        ... )
    """
    allowed = ("dbscan", "adbscan", "optics", "hdbscan")
    if method not in allowed:
        raise ValueError(
            f"Illegal clustering method was given. Got: {method}, allowed: {allowed}"
        )

    if isinstance(gdf, gpd.GeoDataFrame):
        xy = np.vstack([gdf.centroid.x, gdf.centroid.y]).T
    else:
        raise ValueError(
            "The input `gdf` needs to be a gpd.GeoDataFrame with geometry col "
            f"Got: {type(gdf)}"
        )

    if method == "adbscan":
        try:
            from esda.adbscan import ADBSCAN
        except ImportError:
            raise ImportError(
                "The adbscan method requires the esda package to be installed."
                "Install it with: pip install esda"
            )
        xy = pd.DataFrame({"X": xy[:, 0], "Y": xy[:, 1]})
        clusterer = ADBSCAN(eps=eps, min_samples=min_samples, n_jobs=n_jobs, **kwargs)
    elif method == "dbscan":
        clusterer = DBSCAN(eps=eps, min_samples=min_samples, n_jobs=n_jobs, **kwargs)
    elif method == "hdbscan":
        clusterer = HDBSCAN(min_samples=min_samples, n_jobs=n_jobs, **kwargs)
    elif method == "optics":
        clusterer = OPTICS(
            max_eps=eps, min_samples=min_samples, n_jobs=n_jobs, **kwargs
        )

    labels = clusterer.fit(xy).labels_
    gdf["labels"] = labels

    return gdf



from typing import Dict

import geopandas as gpd
import numpy as np
import rasterio
import shapely
from rasterio.features import geometry_mask

__all__ = ["gdf2instance", "gdf2semantic"]


def gdf2instance(gdf: gpd.GeoDataFrame, reset_index: bool = False) -> gpd.GeoDataFrame:
    """Converts a GeoDataFrame to an instance segmentation mask.

    Parameters:
        gdf (gpd.GeoDataFrame):
            GeoDataFrame with a "class_name" column.

    Returns:
        np.ndarray:
            Instance segmentation mask of the input gdf.
    """
    xmin, ymin, xmax, ymax = gdf.total_bounds
    width = xmax - xmin
    height = ymax - ymin
    image_shape = (int(height), int(width))
    out_mask = np.zeros(image_shape, dtype=np.int32)
    for i, (ix, row) in enumerate(gdf.iterrows()):
        mask = geometry_mask(
            [shapely.affinity.translate(row.geometry, -xmin, -ymin)],
            out_shape=image_shape,
            transform=rasterio.Affine(1, 0, 0, 0, 1, 0),
            invert=True,
            all_touched=True,
        )
        if reset_index:
            out_mask[mask] = int(i + 1)
        else:
            out_mask[mask] = int(ix)

    return out_mask


def gdf2semantic(
    gdf: gpd.GeoDataFrame, class_dict: Dict[str, int] = None
) -> np.ndarray:
    """Converts a GeoDataFrame to a semantic segmentation mask.

    Parameters:
        gdf (gpd.GeoDataFrame):
            GeoDataFrame with a "class_name" column.
        class_dict (Dict[str, int]):
            Dictionary mapping class names to integers. If None, the classes
            will be mapped to integers in the order they appear in the GeoDataFrame.

    Returns:
        np.ndarray:
            Semantic segmentation mask of the input gdf.
    """
    xmin, ymin, xmax, ymax = gdf.total_bounds
    width = xmax - xmin
    height = ymax - ymin
    image_shape = (int(height), int(width))
    out_mask = np.zeros(image_shape, dtype=np.int32)
    for i, (cl, gdf) in enumerate(gdf.explode(index_parts=True).groupby("class_name")):
        mask = geometry_mask(
            gdf.geometry.translate(-xmin, -ymin),
            out_shape=image_shape,
            transform=rasterio.Affine(1, 0, 0, 0, 1, 0),
            invert=True,
            all_touched=True,
        )
        if class_dict is None:
            out_mask[mask] = int(i + 1)
        else:
            out_mask[mask] = class_dict[cl]

    return out_mask



from functools import partial
from typing import Any, Callable, Tuple, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely
from shapely.geometry import Polygon, box, mapping

from cellseg_gsontools.apply import gdf_apply
from cellseg_gsontools.utils import get_holes, lonlat_to_xy, xy_to_lonlat

try:
    import h3

    _has_h3 = True
except ImportError:
    _has_h3 = False

__all__ = [
    "bounding_box",
    "get_grid",
    "grid_overlay",
    "hexgrid_overlay",
    "fit_spatial_grid",
]


def fit_spatial_grid(
    gdf: gpd.GeoDataFrame, grid_type: str = "square", **kwargs
) -> gpd.GeoDataFrame:
    """Quick wrapper to fit either a hex or square grid to a `geopandas.GeoDataFrame`.

    Note:
        - Hexagonal grid requires the `h3` package to be installed.
        - Hexagonal grid only works for a gdf containing one single polygon.

    Parameters:
        gdf (gpd.GeoDataFrame):
            GeoDataFrame to fit grid to.
        grid_type (str):
            Type of grid to fit, by default "square".
        **kwargs (Dict[str, Any]):
            Keyword arguments to pass to grid fitting functions.

    Returns:
        gpd.GeoDataFrame:
            Fitted grid.

    Raises:
        ValueError: If grid_type is not one of "square" or "hex".
        ImportError: If grid_type is "hex" and the `h3` package is not installed.

    Examples:
        Fit a hexagonal grid to a gdf:
        >>> from cellseg_gsontools import read_gdf
        >>> from cellseg_gsontools.grid import fit_spatial_grid
        >>> # Read in the tissue areas
        >>> area_gdf = gpd.read_file("path/to/area.geojson")
        >>> # Fit the grid
        >>> hex_grid = fit_spatial_grid(area_gdf, grid_type="hex", resolution=9)
        >>> hex_grid
        gpd.GeoDataFrame

        Fit a square grid to a gdf:
        >>> from cellseg_gsontools import read_gdf
        >>> from cellseg_gsontools.grid import fit_spatial_grid
        >>> # Read in the tissue areas
        >>> area_gdf = gpd.read_file("path/to/area.geojson")
        >>> # Fit the grid
        >>> sq_grid = fit_spatial_grid(
        ...     area_gdf, grid_type="square", patch_size=(256, 256), stride=(256, 256)
        ... )
        >>> sq_grid
        gpd.GeoDataFrame
    """
    allowed = ["square", "hex"]
    if grid_type not in allowed:
        raise ValueError(f"grid_type must be one of {allowed}, got {grid_type}")

    if grid_type == "square":
        grid = grid_overlay(gdf, **kwargs)
    else:
        if not _has_h3:
            raise ImportError("h3 package not installed. Install with `pip install h3`")
        grid = hexgrid_overlay(gdf, **kwargs)

    return grid


def hexgrid_overlay(
    gdf: gpd.GeoDataFrame, resolution: int = 9, to_lonlat: bool = True
) -> gpd.GeoDataFrame:
    """Fit a `h3` hexagonal grid on top of a `geopandas.GeoDataFrame`.

    Parameters:
        gdf (gpd.GeoDataFrame):
            GeoDataFrame to fit grid to.
        resolution (int):
            H3 resolution, by default 9.
        to_lonlat (bool):
            Whether to convert to lonlat coordinates, by default True.

    Returns:
        gpd.GeoDataFrame:
            Fitted h3 hex grid.

    Examples:
        Fit a hexagonal grid to a gdf:
        >>> from cellseg_gsontools import read_gdf
        >>> from cellseg_gsontools.grid import hexgrid_overlay
        >>> # Read in the tissue areas
        >>> area_gdf = gpd.read_file("path/to/area.geojson")
        >>> # Fit the grid
        >>> hex_grid = hexgrid_overlay(area_gdf, resolution=9)
        >>> hex_grid
        gpd.GeoDataFrame
    """
    if gdf.empty or gdf is None:
        return

    # drop invalid geometries if there are any after buffer
    gdf.geometry = gdf.geometry.buffer(0)
    gdf = gdf[gdf.is_valid]

    orig_crs = gdf.crs

    poly = shapely.force_2d(gdf.unary_union)
    if isinstance(poly, Polygon):
        hexagons = poly2hexgrid(poly, resolution=resolution, to_lonlat=to_lonlat)
    else:
        output = []
        for geom in poly.geoms:
            hexes = poly2hexgrid(geom, resolution=resolution, to_lonlat=to_lonlat)
            output.append(hexes)
        hexagons = pd.concat(output)

    return hexagons.set_crs(
        orig_crs, inplace=True, allow_override=True
    ).drop_duplicates("geometry")


def grid_overlay(
    gdf: gpd.GeoDataFrame,
    patch_size: Tuple[int, int] = (256, 256),
    stride: Tuple[int, int] = (256, 256),
    pad: int = 20,
    predicate: str = "intersects",
) -> gpd.GeoDataFrame:
    """Overlay a square grid to the given areas of a `geopandas.GeoDataFrame`.

    Note:
        Returns None if the gdf is empty.

    Parameters:
        gdf (gpd.GeoDataFrame):
            GeoDataFrame to fit the grid to. Uses the bounding box of the GeoDataFrame
            to fit the grid.
        patch_size (Tuple[int, int]):
            Patch size of the grid.
        stride (Tuple[int, int]):
            Stride of the sliding window in the grid.
        pad (int):
            Pad the bounding box with the given number of pixels, by default None.
        predicate (str):
            Predicate to use for the spatial join, by default "intersects".
            Allowed values are "intersects" and "within".

    Returns:
        gpd.GeoDataFrame:
            GeoDataFrame with the grid fitted to the given GeoDataFrame.

    Raises:
        ValueError: If predicate is not one of "intersects" or "within".

    Examples:
        Fit a square grid to a gdf:
        >>> from cellseg_gsontools import read_gdf
        >>> from cellseg_gsontools.grid import grid_overlay
        >>> # Read in the tissue areas
        >>> area_gdf = gpd.read_file("path/to/area.geojson")
        >>> # Fit the grid
        >>> sq_grid = grid_overlay(area_gdf, patch_size=(256, 256), stride=(256, 256))
        >>> sq_grid
        gpd.GeoDataFrame
    """
    if gdf.empty or gdf is None:
        return

    allowed = ["intersects", "within"]
    if predicate not in allowed:
        raise ValueError(f"predicate must be one of {allowed}. Got {predicate}")
    grid = get_grid(gdf, patch_size, stride, pad=pad)
    grid.set_crs(epsg=4328, inplace=True, allow_override=True)
    _, grid_inds = grid.sindex.query(gdf.geometry, predicate=predicate)
    grid = grid.iloc[np.unique(grid_inds)]
    # grid = grid.sjoin(gdf, predicate=predicate)

    return grid.drop_duplicates("geometry")


def poly2hexgrid(
    poly: Polygon, resolution: int = 9, to_lonlat: bool = True
) -> gpd.GeoDataFrame:
    """Convert a shapely Polygon to a h3 hexagon grid.

    Parameters:
        poly (Polygon):
            Polygon to convert.
        resolution (int, optional):
            H3 resolution, by default 9.
        to_lonlat (bool, optional):
            Whether to convert to lonlat coordinates, by default True.

    Returns:
        gpd.GeoDataFrame:
            GeoDataFrame of h3 hexagons.
    """
    x, y = poly.exterior.coords.xy
    if to_lonlat:
        x, y = xy_to_lonlat(x, y)
    holes = get_holes(poly, to_lonlat=to_lonlat)

    poly = Polygon(list(zip(x, y)), holes=holes)
    hexs = h3.polyfill(mapping(poly), resolution, geo_json_conformant=True)

    # to gdf
    hex_polys = gpd.GeoSeries(list(map(polygonise, hexs)), index=hexs, crs=4326)
    hex_polys = gpd.GeoDataFrame(geometry=hex_polys)

    return hex_polys


def polygonise(hex_id: str, to_cartesian: bool = True) -> Polygon:
    """Polygonise a h3 hexagon.

    Parameters:
        hex_id (str):
            H3 hexagon id.
        to_cartesian (bool, optional):
            Whether to convert to cartesian coordinates, by default True.

    Returns:
        Polygon:
            Polygonised h3 hexagon.
    """
    poly = Polygon(h3.h3_to_geo_boundary(hex_id, geo_json=True))

    if to_cartesian:
        lon, lat = poly.exterior.coords.xy
        x, y = lonlat_to_xy(lon, lat)
        poly = Polygon(list(zip(x, y)))

    return poly


def bounding_box(gdf: gpd.GeoDataFrame, pad: int = 0) -> gpd.GeoDataFrame:
    """Get the bounding box of a GeoDataFrame.

    Note:
        returns None if the gdf is empty.

    Parameters:
        gdf (gpd.GeoDataFrame):
            The input GeoDataFrame.
        pad (int):
            The padding to add to the bounding box.

    Returns:
        gpd.GeoDataFrame:
            The bounding box as a GeoDataFrame.
    """
    if gdf.empty or gdf is None:
        return

    xmin, ymin, xmax, ymax = gdf.total_bounds
    bbox = box(xmin - pad, ymin - pad, xmax + pad, ymax + pad)
    return gpd.GeoDataFrame({"geometry": bbox}, index=[0])


def _get_margins(
    first_endpoint: int, size: int, stride: int, pad: int = None
) -> Tuple[int, int]:
    """Get the number of slices needed for one direction and the overlap.

    Parameters:
        first_endpoint (int):
            The first coordinate of the patch.
        size (int):
            The size of the input.
        stride (int):
            The stride of the sliding window.
        pad (int):
            The padding to add to the patch

    Returns:
        Tuple[int, int]:
            The number of patches needed for one direction and the overlap.
    """
    pad = int(pad) if pad is not None else 20  # at least some padding needed
    size += pad

    n = 1
    mod = 0
    end = first_endpoint
    while True:
        n += 1
        end += stride

        if end > size:
            mod = end - size
            break
        elif end == size:
            break

    return n, mod + pad


def get_grid(
    gdf: gpd.GeoDataFrame,
    patch_size: Tuple[int, int] = (256, 256),
    stride: Tuple[int, int] = (256, 256),
    pad: int = 20,
) -> gpd.GeoDataFrame:
    """Get a grid of patches from a GeoDataFrame.

    Note:
        returns None if the gdf is empty.

    Parameters:
        gdf (gpd.GeoDataFrame):
            The input GeoDataFrame.
        patch_size (Tuple[int, int]):
            The size of the patch.
        stride (Tuple[int, int]):
            The stride of the sliding window.
        pad (int):
            The padding to add to the patch

    Returns:
        gpd.GeoDataFrame:
            The grid of patches.
    """
    if gdf.empty or gdf is None:
        return

    # add some default padding
    if pad is None:
        pad = 20

    bbox: gpd.GeoDataFrame = bounding_box(gdf, pad=pad)
    minx, miny, maxx, maxy = bbox.geometry.bounds.values[0]
    width = maxx - minx
    height = maxy - miny

    total_size = (height, width)

    y_end, x_end = patch_size
    nrows, _ = _get_margins(y_end, total_size[0], stride[0], pad=pad)
    ncols, _ = _get_margins(x_end, total_size[1], stride[1], pad=pad)

    grid_rects = []
    for row in range(nrows):
        for col in range(ncols):
            y_start = row * stride[0] + miny
            y_end = y_start + patch_size[0]
            x_start = col * stride[1] + minx
            x_end = x_start + patch_size[1]
            rect = Polygon(
                [(x_start, y_start), (x_end, y_start), (x_end, y_end), (x_start, y_end)]
            )
            grid_rects.append(rect)

    return gpd.GeoDataFrame({"geometry": grid_rects})


def get_rect_metric(
    rect, objs: gpd.GeoDataFrame, metric_func: Callable, predicate: str
) -> Any:
    """Get the metric of the given rectangle.

    Parameters:
        rect (Polygon):
            The rectangle to get the metric of.
        objs (gpd.GeoDataFrame):
            The objects to use for the metric.
        metric_func (Callable):
            The metric function to use.
        predicate (str):
            The predicate to use for the spatial join. Allowed values are "intersects"
            and "within".

    Returns:
        Any:
            The metric of the rectangle.
    """
    allowed = ["intersects", "within"]
    if predicate not in allowed:
        raise ValueError(f"predicate must be one of {allowed}. Got {predicate}.")

    sub_objs = objs.iloc[objs.geometry.sindex.query(rect, predicate=predicate)]

    return metric_func(sub_objs)


def grid_classify(
    grid: gpd.GeoDataFrame,
    objs: gpd.GeoDataFrame,
    metric_func: Callable,
    predicate: str,
    new_col_names: Union[Tuple[str, ...], str],
    parallel: bool = True,
    num_processes: int = -1,
    pbar: bool = False,
    **kwargs,
) -> gpd.GeoDataFrame:
    """Classify the grid based on objs inside the grid cells.

    Parameters:
        grid (gpd.GeoDataFrame):
            The grid of rectangles to classify.
        objs (gpd.GeoDataFrame):
            The objects to use for classification.
        metric_func (Callable):
            The metric/heuristic function to use for classification.
        predicate (str):
            The predicate to use for the spatial join. Allowed values are "intersects"
            and "within".
        new_col_names (Union[Tuple[str, ...], str]):
            The name of the new column(s) in the grid gdf.
        parallel (bool):
            Whether to use parallel processing.
        num_processes (int):
            The number of processes to use. If -1, uses all available cores.
            Ignored if parallel=False.
        pbar (bool):
            Whether to show a progress bar. Ignored if parallel=False.

    Returns:
        gpd.GeoDataFrame:
            The grid with the new columns added.

    Raises:
        ValueError: If predicate is not one of "intersects" or "within".

    Examples:
        Get the number of immune cells in each grid cell at the tumor stroma interface:
        >>> import geopandas as gpd
        >>> from shapely.geometry import Polygon
        >>> from functools import partial
        >>> from cellseg_gsontools import gdf_apply, read_gdf
        >>> from cellseg_gsontools.grid import grid_classify, grid_overlay
        >>> from cellseg_gsontools.context import InterfaceContext
        >>> # Define a heuristic function to get the number of immune cells
        >>> def get_immune_cell_cnt(gdf: gpd.GeoDataFrame, **kwargs) -> int:
        ...     try:
        ...         cnt = gdf.class_name.value_counts()["inflammatory"]
        ...     except KeyError:
        ...         cnt = 0
        ...     return int(cnt)
        >>> # Read in the tissue areas and cells
        >>> area_gdf = gpd.read_file("path/to/area.geojson")
        >>> cell_gdf = gpd.read_file("path/to/cell.geojson")
        >>> # Fit a tumor-stroma interface
        >>> tumor_stroma_iface = InterfaceContext(
        ...     area_gdf=area_gdf,
        ...     cell_gdf=cell_gdf,
        ...     top_labels="area_cin",
        ...     bottom_labels="areastroma",
        ...     buffer_dist=250,
        ...     graph_type="distband",
        ...     dist_thresh=75,
        ...     patch_size=(128, 128),
        ...     stride=(128, 128),
        ...     min_area_size=50000,
        ... )
        >>> tumor_stroma_iface.fit(verbose=False)
        >>> # Get the grid and the cells at the interface
        >>> iface_grid = grid_overlay(
        ...     tumor_stroma_iface.context2gdf("interface_area"),
        ...     patch_size=(128, 128),
        ...     stride=(128, 128),
        ... )
        >>> cells = tumor_stroma_iface.context2gdf("interface_cells")
        >>> # Classify the grid
        >>> iface_grid = grid_classify(
        >>>     grid=iface_grid,
        >>>     objs=cells,
        >>>     metric_func=get_immune_cnt,
        >>>     predicate="intersects",
        >>>     new_col_name="immune_cnt",
        >>>     parallel=True,
        >>>     pbar=True,
        >>>     num_processes=-1
        >>> )
        >>> iface_grid
        geometry  immune_cnt
        28  POLYGON ((20032.00000 54098.50000, 20160.00000... 15
        29  POLYGON ((20160.00000 54098.50000, 20288.00000... 3
    """
    allowed = ["intersects", "within"]
    if predicate not in allowed:
        raise ValueError(f"predicate must be one of {allowed}. Got {predicate}")

    if isinstance(new_col_names, str):
        new_col_names = [new_col_names]

    func = partial(
        get_rect_metric, objs=objs, predicate=predicate, metric_func=metric_func
    )
    grid.loc[:, list(new_col_names)] = gdf_apply(
        grid,
        func=func,
        parallel=parallel,
        pbar=pbar,
        num_processes=num_processes,
        columns=["geometry"],
    )

    return grid



from functools import partial
from itertools import combinations_with_replacement
from typing import Dict, List, Tuple

import geopandas as gpd
from libpysal.weights import W
from shapely.geometry import LineString, Point

from .apply import gdf_apply

__all__ = ["weights2gdf", "link_counts", "get_link_combinations"]


def get_link_combinations(classes: Tuple[str, ...]) -> List[str]:
    """Return a list of link combinations between the classes in `classes`.

    Parameters:
        classes (Tuple[str, ...]):
            A list/tuple containing the classes of your dataset.
    """
    combos = ["-".join(t) for t in list(combinations_with_replacement(classes, 2))]

    return combos


def _create_link(focal: Point, neighbor: Point) -> LineString:
    """Create a LineString between two centroids.

    Parameters:
        focal (Point):
            Focal centroid.
        neighbor (Point):
            Neighbor centroid.

    Returns:
        LineString:
            LineString between the two centroids.
    """
    return LineString([focal, neighbor])


def _get_link_class(
    focal_class: str, neighbor_class: str, link_combos: List[str]
) -> str:
    """Get link class based on focal and neighbor class.

    Parameters:
        focal_class (str):
            Focal class name.
        neighbor_class (str):
            Neighbor class name.
        link_combos (List[str]):
            List of all possible link class combinations.

    Returns:
        str:
            Link class name.
    """
    for link_class in link_combos:
        class1, class2 = link_class.split("-")
        if (focal_class == class1 and neighbor_class == class2) or (
            focal_class == class2 and neighbor_class == class1
        ):
            return link_class
    return None


def weights2gdf(
    gdf: gpd.GeoDataFrame, w: W, parallel: bool = False
) -> gpd.GeoDataFrame:
    """Convert a `libpysal` weights object to a `geopandas.GeoDataFrame`.

    Add class names and node centroids to the dataframe.

    Note:
        if `w.neighbors` is empty, this will return None.

    Parameters:
        gdf (gpd.GeoDataFrame):
            GeoDataFrame of the nodes.
        w (W):
            PySAL weights object.
        parallel (bool, default=False):
            Whether to use parallel processing.

    Returns:
        gpd.GeoDataFrame:
            GeoDataFrame of the links.

    Examples:
        Convert `libpysal` weights from `InterfaceContext` to `geopandas.GeoDataFrame`.
        >>> from cellseg_gsontools.spatial_context import InterfaceContext
        >>> from cellseg_gsontools.links import weights2gdf
        >>> iface_context = InterfaceContext(
        ...     area_gdf=areas,
        ...     cell_gdf=cells,
        ...     top_labels="area_cin",
        ...     bottom_labels="areastroma",
        ...     silence_warnings=True,
        ...     min_area_size=100000.0,
        ... )
        >>> iface_context.fit(verbose=False)
        >>> w = iface_context.context2weights("border_network")
        >>> link_gdf = weights2gdf(cells, w)
    """
    if not w.neighbors:
        return

    # get all possible link class combinations
    classes = sorted(gdf.class_name.unique().tolist())
    link_combos = get_link_combinations(classes)

    # init link gdf
    link_gdf = w.to_adjlist(remove_symmetric=True, drop_islands=True).reset_index()

    # add centroids and class names
    link_gdf.loc[:, "focal_centroid"] = gdf.loc[link_gdf.focal].centroid.to_list()
    link_gdf.loc[:, "neighbor_centroid"] = gdf.loc[link_gdf.neighbor].centroid.to_list()
    link_gdf.loc[:, "focal_class_name"] = gdf.loc[link_gdf.focal].class_name.to_list()
    link_gdf.loc[:, "neighbor_class_name"] = gdf.loc[
        link_gdf.neighbor
    ].class_name.to_list()

    func = partial(_get_link_class, link_combos=link_combos)
    link_gdf["class_name"] = gdf_apply(
        link_gdf,
        func=func,
        columns=["focal_class_name", "neighbor_class_name"],
        axis=1,
        parallel=parallel,
    )

    link_gdf["geometry"] = gdf_apply(
        link_gdf,
        func=_create_link,
        columns=["focal_centroid", "neighbor_centroid"],
        axis=1,
        parallel=parallel,
    )
    link_gdf = link_gdf.set_geometry("geometry")

    return link_gdf


def link_counts(
    gdf: gpd.GeoDataFrame, w: W, classes: Tuple[str, ...]
) -> Dict[str, int]:
    """Get the link-type counts of a geodataframe given a spatial weights object `w`.

    Parameters:
        gdf (gpd.GeoDataFrame):
            Input geodataframe. Has to have a `class_name` column.
        w (libysal.weights.W):
            Libpysal spatial weights object of the gdf.
        classes (Tuple[str, ...]):
            A list/tuple containing the classes of your dataset.

    Returns:
        Dict[str, int]:
            A contigency dictionary.

    Examples:
        Get the link types of the tumor-stroma interfaces.
        >>> from cellseg_gsontools.spatial_context import InterfaceContext
        >>> from cellseg_gsontools.links import link_counts
        >>> iface_context = InterfaceContext(
        ...     area_gdf=areas,
        ...     cell_gdf=cells,
        ...     top_labels="area_cin",
        ...     bottom_labels="areastroma",
        ...     silence_warnings=True,
        ...     min_area_size=100000.0,
        ... )
        >>> classes = [
        ...     "inflammatory",
        ...     "connective",
        ...     "glandular_epithel",
        ...     "squamous_epithel",
        ...     "neoplastic",
        ... ]
        >>> iface_context.fit(verbose=False)
        >>> w = iface_context.context2weights("border_network")
        >>> link_counts(cells, w, classes)
        {'inflammatory-inflammatory': 31,
        'inflammatory-connective': 89,
        'inflammatory-glandular_epithel': 0,
        'inflammatory-squamous_epithel': 0,
        'inflammatory-neoplastic': 86,
        'connective-connective': 131,
        'connective-glandular_epithel': 0,
        'connective-squamous_epithel': 0,
        'connective-neoplastic': 284,
        'glandular_epithel-glandular_epithel': 0,
        'glandular_epithel-squamous_epithel': 0,
        'glandular_epithel-neoplastic': 0,
        'squamous_epithel-squamous_epithel': 0,
        'squamous_epithel-neoplastic': 0,
        'neoplastic-neoplastic': 236}
    """
    combos = get_link_combinations(classes)
    link_cnt = {combo: 0 for combo in combos}

    all_node_pairs = []
    for node, neighbors in w.neighbors.items():
        ncls = gdf.loc[node, "class_name"]
        if neighbors:
            neighbors_cls = gdf.loc[neighbors, "class_name"]
            node_pairs = [set((node, n)) for n in neighbors]

            for ngh_ix, ngh_cls in zip(neighbors, neighbors_cls):
                node_ids = set((node, ngh_ix))
                if node_ids not in all_node_pairs:
                    for combo in combos:
                        types = set(combo.split("-"))
                        if ncls in types and ngh_cls in types and ngh_cls != ncls:
                            link_cnt[combo] += 1
                        elif ncls in types and ngh_cls in types and ngh_cls == ncls:
                            if len(types) == 1:
                                link_cnt[combo] += 1

            all_node_pairs.extend(node_pairs)

    return link_cnt



from collections import defaultdict
from functools import partial
from typing import Any, Tuple

import geopandas as gpd
import numpy as np
import shapely
from libpysal.weights import W
from scipy.spatial import Voronoi
from shapely import vectorized

from cellseg_gsontools.apply import gdf_apply
from cellseg_gsontools.geometry import major_axis_len

__all__ = [
    "perpendicular_line",
    "equal_interval_points",
    "voronoi_medial",
    "line_branches",
    "medial_lines",
    "perpendicular_lines",
]


def perpendicular_line(
    line: shapely.LineString, seg_length: float
) -> shapely.LineString:
    """Create a perpendicular line from a line segment.

    Note:
        Returns an empty line if perpendicular line is not possible from the input.

    Parameters:
        line (shapely.LineString):
            Line segment to create a perpendicular line from.
        seg_length (float):
            Length of the perpendicular line.

    Returns:
        shapely.LineString:
            Perpendicular line to the input line of length `seg_length`.
    """
    left = line.parallel_offset(seg_length / 2, "left").centroid
    right = line.parallel_offset(seg_length / 2, "right").centroid

    if left.is_empty or right.is_empty:
        return shapely.LineString()

    return shapely.LineString([left, right])


def equal_interval_points(obj: Any, n: int = None, delta: float = None):
    """Resample the points of a shapely object at equal intervals.

    Parameters:
        obj (Any):
            Any shapely object that has length property.
        n (int):
            Number of points, defaults to None
        delta (float):
            Distance between points, defaults to None

    Returns:
        points (numpy.ndarray):
            Array of points at equal intervals along the input object.
    """
    length = obj.length

    if n is None:
        if delta is None:
            delta = obj.length / 1000
        n = round(length / delta)

    distances = np.linspace(0, length, n)
    points = [obj.interpolate(distance) for distance in distances]
    points = np.array([(p.x, p.y) for p in points])

    return points


# Adapted from
# https://github.com/mikedh/trimesh/blob/main/trimesh/path/polygons.py
def voronoi_medial(
    polygon: shapely.Polygon, n: int = None, delta: float = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Compute the medial lines of a polygon using voronoi diagram.

    Parameters:
        polygon (shapely.geometry.Polygon):
            Polygon to compute the medial lines of.
        n (int):
            Number of resampled points in the input polygon, defaults to None
        delta (float):
            Distance between resampled polygon points, defaults to None. Ignored
            if n is not None.

    Returns:
        vertices (numpy.ndarray):
            Array of vertices of the voronoi diagram.
        edges (numpy.ndarray):
            Array of edges of the voronoi diagram.
    """
    points = equal_interval_points(polygon.exterior, n=n, delta=delta)

    # # create the voronoi diagram on 2D points
    voronoi = Voronoi(points)

    # which voronoi vertices are contained inside the polygon
    contains = vectorized.contains(polygon, *voronoi.vertices.T)

    # ridge vertices of -1 are outside, make sure they are False
    contains = np.append(contains, False)

    # make sure ridge vertices is numpy array
    ridge = np.asanyarray(voronoi.ridge_vertices, dtype=np.int64)

    # only take ridges where every vertex is contained
    edges = ridge[contains[ridge].all(axis=1)]

    # now we need to remove uncontained vertices
    contained = np.unique(edges)
    mask = np.zeros(len(voronoi.vertices), dtype=np.int64)
    mask[contained] = np.arange(len(contained))

    # mask voronoi vertices
    vertices = voronoi.vertices[contained]

    # re-index edges
    return vertices, mask[edges]


def line_branches(edges: np.ndarray) -> np.ndarray:
    """Get the branch points of a line graph.

    Note:
        Helps to get rid of the random branches of the voronoi medial lines.

    Parameters:
        edges (numpy.ndarray):
            Array of edges of the line graph.

    Returns:
        numpy.ndarray:
            Array of edges of the line graph with the branches removed.
    """
    # create a graph from the edges
    neighbors = defaultdict(list)
    for line in edges:
        start_id, end_id = line
        neighbors[start_id].append(end_id)
        neighbors[end_id].append(start_id)

    w = W(dict(sorted(neighbors.items())))

    # get the branch points
    branch_points = [k for k, c in w.cardinalities.items() if c > 2]

    # get the paths from the branch points
    paths = []
    stack = [(bp, None, [bp]) for bp in branch_points]
    while stack:
        cur, prev, path = stack.pop()

        if len(w.neighbors[cur]) == 1 or (prev and cur in branch_points):
            paths.append(path)
            continue

        for neighbor in w.neighbors[cur]:
            if neighbor != prev:
                stack.append((neighbor, cur, path + [neighbor]))

    return paths


def medial_lines(
    polygon: shapely.Polygon,
    n: int = None,
    delta: float = None,
    rm_branches: bool = False,
) -> gpd.GeoDataFrame:
    """Get the medial lines of a polygon using a voronoi diagram.

    Parameters:
        polygon (shapely.Polygon):
            Polygon to compute the medial lines of.
        n (int):
            Number of resampled points in the input polygon, defaults to None
        delta (float):
            Distance between resampled polygon points, defaults to None. Ignored
            if n is not None.
        rm_branches (bool):
            Whether to remove the branches of the medial lines, defaults to False.

    Returns:
        gpd.GeoDataFrame:
            GeoDataFrame of the medial lines.

    Examples:
        >>> import geopandas as gpd
        >>> import shapely
        >>> from cellseg_gsontools.lines import medial_lines
        >>> from cellseg_gsontools.data import cervix_tissue
        >>> tissues = cervix_tissue()
        >>> tumor = tissues[tissues["class_name"] == "area_cin"]
        >>> tumor_poly = tumor.geometry.iloc[0]  # get the first tumor polygon
        >>> polygon = shapely.Polygon(tumor_poly.exterior)
        >>> med_lines = medial_lines(polygon, delta=500, rm_branches=False)
        >>> med_lines.head()
        geometry
            0  LINESTRING (10789.887 49746.299, 10910.622 493...
            1  LINESTRING (10789.887 49746.299, 10926.865 498...
            2  LINESTRING (10924.971 48929.809, 10829.145 492...
            3  LINESTRING (10910.622 49332.471, 10829.145 492...
            4  LINESTRING (10926.865 49843.003, 10794.602 502...
    """
    # get the medial lines
    vertices, edges = voronoi_medial(polygon, n=n, delta=delta)

    # remove lone branches of the medial lines
    if rm_branches:
        # get the line paths and branches of the medial lines
        paths = line_branches(edges)
        edges = np.vstack(
            [
                np.array(list(zip(branch, branch[1:])))
                for branch in paths
                if len(branch) > 2
            ]
        )

    med_lines = gpd.GeoDataFrame(
        [shapely.LineString(vertices[line]) for line in edges], columns=["geometry"]
    )
    # clip the medial lines to the polygon
    med_lines = med_lines.loc[med_lines.within(polygon)]

    return med_lines


def perpendicular_lines(
    lines: gpd.GeoDataFrame, polygon: shapely.Polygon = None
) -> gpd.GeoDataFrame:
    """Get perpendicular lines to the input lines starting from the line midpoints.

    Parameters:
        lines (gpd.GeoDataFrame):
            GeoDataFrame of the input lines.
        polygon (shapely.Polygon):
            Polygon to clip the perpendicular lines to.

    Returns:
        gpd.GeoDataFrame:
            GeoDataFrame of the perpendicular lines.
    """
    # create perpendicular lines to the medial lines
    if polygon is None:
        polygon = lines.unary_union.convex_hull

    seg_len = major_axis_len(polygon)
    func = partial(perpendicular_line, seg_length=seg_len)
    perp_lines = gdf_apply(lines, func, columns=["geometry"])

    # clip the perpendicular lines to the polygon
    perp_lines = gpd.GeoDataFrame(perp_lines, columns=["geometry"]).clip(polygon)

    # explode perpendicular lines & take only the ones that intersect w/ medial lines
    perp_lines = perp_lines.explode(index_parts=False).reset_index(drop=True)

    # drop the perpendicular lines that are too short or too long
    # since these are likely artefacts
    perp_lines["len"] = perp_lines.geometry.length
    low, high = perp_lines.len.quantile([0.05, 0.85])
    perp_lines = perp_lines.query(f"{low}<len<{high}")

    return perp_lines



from .fetch import (
    cell_merge_dir,
    cervix_cells,
    cervix_tissue,
    gland_cells,
    gland_tissue,
    tissue_merge_dir,
    tumor_stroma_intreface_cells,
    tumor_stroma_intreface_tissue,
)

__all__ = [
    "cervix_cells",
    "cervix_tissue",
    "gland_cells",
    "gland_tissue",
    "tumor_stroma_intreface_cells",
    "tumor_stroma_intreface_tissue",
    "cell_merge_dir",
    "tissue_merge_dir",
]



# try:
#     import dask_geopandas

#     _has_dask_geopandas = True
# except ImportError:
#     _has_dask_geopandas = False


# import geopandas as gpd

# from cellseg_gsontools.spatial_context.backend import _SpatialBackend
# from cellseg_gsontools.spatial_context.ops import (
#     get_interface_zones,
#     get_objs,
#     get_objs_dgp,
# )

# __all__ = ["_SpatialContextDGP"]


# class _SpatialContextDGP(_SpatialBackend):
#     def __init__(self) -> None:
#         """Create a dask_geopandas spatial context class."""
#         if not _has_dask_geopandas:
#             raise ImportError(
#                 "dask_geopandas is required to use the spatial context backend."
#                 "Install it with `pip install dask_geopandas`."
#             )
#         self.backend_name = "dask_geopandas"

#     def roi(self, context_area: gpd.GeoDataFrame, ix: int) -> gpd.GeoDataFrame:
#         """Get a roi area of index `ix`.

#         Parameters
#         ----------
#         context_area : gpd.GeoDataFrame
#             The context area gdf.
#         ix : int
#             The index of the roi area. I.e., the ith roi area.

#         Returns
#         -------
#             gpd.GeoDataFrame:
#                 The ith roi area.
#         """
#         row: gpd.GeoSeries = context_area.loc[ix]
#         roi_area = gpd.GeoDataFrame([row], crs=context_area.crs)

#         return roi_area

#     def roi_cells(
#         self,
#         roi_area: gpd.GeoDataFrame,
#         cell_gdf: gpd.GeoDataFrame,
#         cell_gdf_dgp: dask_geopandas.GeoDataFrame = None,
#         predicate: str = "intersects",
#         silence_warnings: bool = True,
#         parallel: bool = False,
#         **kwargs,
#     ) -> gpd.GeoDataFrame:
#         """Get the cells within the roi area.

#         Parameters
#         ----------
#         roi_area : gpd.GeoDataFrame
#             The roi area.
#         cell_gdf : gpd.GeoDataFrame
#             The cell gdf.
#         cell_gdf_dgp : dask_geopandas.GeoDataFrame, optional
#             The cell gdf as a dask_geopandas.GeoDataFrame. If None, then the geopandas
#             cell_gdf is used.
#         predicate : str, default="intersects"
#             The predicate to use for the spatial join. See `geopandas.tools.sjoin`
#         silence_warnings : bool, default=True
#             Flag, whether to silence all the geopandas related warnings.
#         parallel : bool, default=False
#             Flag, whether to parallelize the spatial join.

#         Returns
#         -------
#             gpd.GeoDataFrame:
#                 The cells within the roi area.

#         Returns
#         -------
#             gpd.GeoDataFrame:
#                 The cells within the roi area.
#         """
#         if roi_area is None or roi_area.empty:
#             return

#         # get the cells within the roi area
#         # if area is large, use dask-geopandas
#         if (
#             roi_area.geometry.area.values[0] > 1e8
#             and parallel
#             and cell_gdf_dgp is not None
#         ):
#             objs_within: gpd.GeoDataFrame = get_objs_dgp(
#                 roi_area,
#                 cell_gdf_dgp,
#                 silence_warnings=silence_warnings,
#                 predicate=predicate,
#             )
#         else:
#             objs_within: gpd.GeoDataFrame = get_objs(
#                 roi_area,
#                 cell_gdf,
#                 silence_warnings=silence_warnings,
#                 predicate=predicate,
#             )

#         if objs_within is None or objs_within.empty:
#             return

#         # rename spatial join columns
#         objs_within = objs_within.rename(
#             columns={
#                 "index_right": "spatial_context_id",
#                 "global_id_left": "global_id",
#                 "class_name_left": "class_name",
#                 "class_name_right": "spatial_context",
#             },
#         )

#         # drop unnecessary columns and return
#         objs_within.drop(columns=["global_id_right"], inplace=True)

#         # convert to geopandas if dask-geopandas
#         if objs_within.index.name == "hilbert_distance":
#             objs_within = self.to_geopandas(objs_within)

#         return objs_within.set_geometry("geometry")

#     def interface(
#         self,
#         top_roi_area: gpd.GeoDataFrame,
#         bottom_gdf: gpd.GeoDataFrame,
#         buffer_dist: int,
#     ) -> gpd.GeoDataFrame:
#         """Get an interface area of index `ix`.

#         Parameters
#         ----------
#             top_roi_area : gpd.GeoDataFrame
#                 The top roi area.
#             bottom_gdf : gpd.GeoDataFrame
#                 The bottom gdf on top of which top_roi_area is buffered.
#         """
#         # Get the intersection of the roi and the area of type `label2`
#         iface = get_interface_zones(
#             buffer_area=top_roi_area,
#             areas=bottom_gdf,
#             buffer_dist=buffer_dist,
#         )

#         # If there are many interfaces, dissolve them into one
#         if len(iface) > 1:
#             iface = iface.dissolve().set_geometry("geometry")

#         return iface

#     def to_geopandas(self, gdf):
#         """Convert gdf to right SpatialContext gdf format."""
#         return gdf.sort_values("global_id").set_index("global_id", drop=False)

#     def convert_area_gdf(self, gdf: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
#         """Convenience identity function to match the backend API.."""
#         return gdf

#     def convert_cell_gdf(
#         self, gdf: gpd.GeoDataFrame, n_partitions: int = None, **kwargs
#     ) -> dask_geopandas.GeoDataFrame:
#         """Set up a dask_geopandas.GeoDataFrame attribute for parallel spatial ops.

#         Parameters
#         ----------
#         gdf : gpd.GeoDataFrame
#             The cell gdf to convert.
#         n_partitions : int, optional
#             The number of processes/partitions to use for the DaskGeoDataFrame. If None,
#             the number of partitions is set to the number of cores.

#         Attributes
#         ----------
#         cell_gdf_dgp : DaskGeoDataFrame
#             The cell gdf as a dask_geopandas.GeoDataFrame.

#         Returns
#         -------
#         gpd.GeoDataFrame:
#             Return the original gpd.GeoDataFrame.
#         """
#         if _has_dask_geopandas:
#             print(
#                 f"partitioning cell_gdf into {n_partitions} partitions and shuffling."
#             )
#             self.cell_gdf_dgp = dask_geopandas.from_geopandas(
#                 gdf, npartitions=n_partitions
#             )
#             self.cell_gdf_dgp = self.cell_gdf_dgp.spatial_shuffle()
#         else:
#             raise ImportError(
#                 "dask-geopandas is required to use the spatial context "
#                 "backend. with backend == 'dask-geopandas'."
#                 "Install with `pip install dask-geopandas`."
#             )

#         return gdf



# from typing import Union

# try:
#     from spatialpandas import GeoDataFrame
#     from spatialpandas.dask import DaskGeoDataFrame

#     _has_spatialpandas = True
# except ImportError:
#     _has_spatialpandas = False

# try:
#     import dask
#     import dask.dataframe as dd
#     from dask.distributed import Client, LocalCluster

#     _has_dask = True
# except ImportError:
#     _has_dask = False

# import geopandas as gpd

# from cellseg_gsontools.spatial_context.ops import get_interface_zones, get_objs_sp

# from ._base_backend import _SpatialBackend

# if _has_dask:
#     dask.config.set({"logging.distributed": "error"})

# __all__ = ["_SpatialContextSP"]


# class _SpatialContextSP(_SpatialBackend):
#     def __init__(self) -> None:
#         """Create a spatialpandas spatial context class."""
#         if not _has_spatialpandas:
#             raise ImportError(
#                 "spatialpandas is required to use the spatial context backend."
#                 "Install it with `pip install spatialpandas`."
#             )
#         self.backend_name = "spatialpandas"

#     def roi(self, context_area: gpd.GeoDataFrame, ix: int) -> gpd.GeoDataFrame:
#         """Get a roi area of index `ix`.

#         Parameters
#         ----------
#         context_area : gpd.GeoDataFrame
#             The context area gdf.
#         ix : int
#             The index of the roi area. I.e., the ith roi area.

#         Returns
#         -------
#             gpd.GeoDataFrame:
#                 The ith roi area.
#         """
#         roi_area = context_area.loc[[ix]]

#         return self.to_geopandas(roi_area)

#     def roi_cells(
#         self,
#         roi_area: gpd.GeoDataFrame,
#         cell_gdf: gpd.GeoDataFrame,
#         predicate: str = "intersects",
#         silence_warnings: bool = True,
#         parallel: bool = False,
#         num_processes: int = None,
#         **kwargs,
#     ) -> gpd.GeoDataFrame:
#         """Get the cells within the roi area.

#         Parameters
#         ----------
#         roi_area : gpd.GeoDataFrame
#             The roi area.
#         cell_gdf : GeoDataFrame
#             The cell gdf.
#         predicate : str, default="intersects"
#             The predicate to use for the spatial join.
#         silence_warnings : bool, default=True
#             Flag, whether to silence all the geopandas related warnings.
#         parallel : bool, default=False
#             Whether to use parallel processing.
#         num_processes : int, optional
#             The number of processes/partitions to use for the DaskGeoDataFrame. If None,
#             the number of partitions is set to the number of cores.

#         Returns
#         -------
#             gpd.GeoDataFrame:
#                 The cells within the roi area.
#         """
#         # check to not compute roi_area if already computed
#         if roi_area is None or roi_area.empty:
#             return

#         # distributing sjoin to many workers results in speedup only for large areas
#         if roi_area.geometry.area.values[0] > 1e8 and parallel:
#             with LocalCluster(
#                 n_workers=int(num_processes),
#                 processes=True,
#                 threads_per_worker=1,
#             ) as cluster, Client(cluster) as client:  # noqa
#                 objs_within: GeoDataFrame = get_objs_sp(
#                     GeoDataFrame(roi_area),
#                     cell_gdf,
#                     silence_warnings=silence_warnings,
#                     predicate=predicate,
#                 )
#         else:
#             objs_within: GeoDataFrame = get_objs_sp(
#                 GeoDataFrame(roi_area),
#                 cell_gdf,
#                 silence_warnings=silence_warnings,
#                 predicate=predicate,
#             )

#         if objs_within is None or len(objs_within.index) == 0:
#             return

#         # rename spatial join columns
#         objs_within = objs_within.rename(
#             columns={
#                 "index_right": "spatial_context_id",
#                 "global_id_left": "global_id",
#                 "class_name_left": "class_name",
#                 "class_name_right": "spatial_context",
#             },
#         )

#         # drop unnecessary columns and return
#         objs_within = objs_within.drop(columns=["global_id_right"])

#         return self.to_geopandas(objs_within)

#     def interface(
#         self,
#         top_roi_area: gpd.GeoDataFrame,
#         bottom_gdf: GeoDataFrame,
#         buffer_dist: int,
#     ) -> gpd.GeoDataFrame:
#         """Get an interface area of index `ix`.

#         Parameters
#         ----------
#             top_roi_area : gpd.GeoDataFrame
#                 The top roi area.
#             bottom_gdf : gpd.GeoDataFrame
#                 The bottom gdf on top of which top_roi_area is buffered.
#         """
#         # Get the intersection of the roi and the area of type `label2`
#         iface = get_interface_zones(
#             buffer_area=top_roi_area,
#             areas=bottom_gdf.to_geopandas().set_crs(
#                 epsg=4328, inplace=True, allow_override=True
#             ),
#             buffer_dist=buffer_dist,
#         )

#         # If there are many interfaces, dissolve them into one
#         if len(iface) > 1:
#             iface = iface.dissolve().set_geometry("geometry")

#         return iface

#     def to_geopandas(self, gdf) -> gpd.GeoDataFrame:
#         """Convert a spatialpandas GeoDataFrame to a geopandas GeoDataFrame."""
#         if isinstance(gdf, GeoDataFrame):
#             gdf = (
#                 gdf.to_geopandas()
#                 .set_geometry("geometry")
#                 .set_crs(epsg=4328, inplace=True, allow_override=True)
#             )
#         return gdf

#     def convert_area_gdf(self, gdf: gpd.GeoDataFrame) -> GeoDataFrame:
#         """Convert area gdf to spatialpandas gdf."""
#         if _has_spatialpandas:
#             if isinstance(gdf, gpd.GeoDataFrame):
#                 gdf = GeoDataFrame(gdf)
#             else:
#                 raise TypeError(
#                     f"input gdf has to be a geopandas.GeoDataFrame. Got: {type(gdf)}"
#                 )
#         else:
#             raise ImportError(
#                 "spatialpandas is required to use the spatial context backend."
#                 "with backend == 'spatialpandas'. Install it with "
#                 "`pip install spatialpandas`."
#             )

#         return gdf

#     def convert_cell_gdf(
#         self, gdf: gpd.GeoDataFrame, parallel: bool, n_partitions: int = None
#     ) -> Union[GeoDataFrame, DaskGeoDataFrame]:
#         """Convert cell gdf to spatialpandas gdf.

#         Parameters
#         ----------
#         gdf : gpd.GeoDataFrame
#             The cell gdf to convert.
#         parallel : bool
#             Whether to convert to a DaskGeoDataFrame.
#         n_partitions : int, optional
#             The number of processes/partitions to use for the DaskGeoDataFrame. If None,
#             the number of partitions is set to the number of cores.

#         Returns
#         -------
#         Union[GeoDataFrame, DaskGeoDataFrame]:
#             The converted cell gdf.
#         """
#         if _has_spatialpandas:
#             if isinstance(gdf, gpd.GeoDataFrame):
#                 # spatialpandas can use only Point arrays in sjoin
#                 gdf["centroid"] = gdf.centroid
#                 gdf = gdf.set_geometry("centroid")
#                 gdf = GeoDataFrame(gdf)
#             else:
#                 raise TypeError(
#                     f"input gdf has to a geopandas.GeoDataFrame. Got: {type(gdf)}"
#                 )
#         else:
#             raise ImportError(
#                 "spatialpandas is required to use the spatial context backend."
#                 "with backend == 'spatialpandas'. Install it with "
#                 "`pip install spatialpandas`"
#             )

#         if parallel:
#             if _has_dask and _has_spatialpandas:
#                 print(
#                     f"partitioning cell_gdf into {n_partitions} partitions. This can "
#                     "take a few minutes...",
#                 )
#                 gdf = dd.from_pandas(gdf, npartitions=n_partitions).persist()
#                 print("partitioning done.")

#                 # Precompute the partition-level spatial index
#                 gdf.partition_sindex
#             else:
#                 raise ImportError(
#                     "spatialpandas and dask is required to use the spatial context "
#                     "backend. with backend == 'spatialpandas' and parallel==True. "
#                     "Install with `pip install spatialpandas` and `pip install dask`."
#                 )
#         return gdf



import warnings
from typing import Optional, Sequence

import geopandas as gpd
import numpy as np
import pandas as pd
from libpysal.weights import KNN, Delaunay, DistanceBand, Relative_Neighborhood, W

__all__ = [
    "fit_graph",
    "dist_thresh_weights_sequential",
    "get_border_crosser_links",
    "_drop_neighbors",
]


def fit_graph(
    gdf: gpd.GeoDataFrame,
    type: str,
    id_col: Optional[str] = None,
    thresh: Optional[float] = None,
    silence_warnings: bool = True,
    **kwargs,
) -> W:
    """Fit a `libpysal` spatial weights graph to a gdf.

    Optionally, a distance threshold can be set for edges that are too long.

    This is a wrapper to fit `libpysal` graph with additional distance threshing.

    Note:
        Allowed graph fitting methods:

        - `delaunay`
        - `knn`
        - `distband`
        - `relative_nhood`

    Parameters:
        gdf (gpd.GeoDataFrame):
            The input geodataframe.
        type (str):
            The type of the libpysal graph. Allowed: "delaunay", "knn", "distband",
            "relative_nhood"
        id_col (str):
            The unique id column in the gdf. If None, this uses `set_uid` to set it.
        thresh (float):
            A distance threshold for too long edges.
        silence_warnings (bool):
            Flag to silence the warnings.
        **kwargs (Dict[str, Any]):
            Arbitrary keyword arguments for the Graph init functions.

    Returns:
        libpysal.weights.W or None:
            A libpysal spatial weights object, containing the neighbor graph data.
            Returns None if the input gdf is empty.

    Examples:
        Fit a DistanceBand to a gdf with a dist threshold of 120.0.
        >>> from cellseg_gsontools.graphs import fit_graph
        >>> w = fit_graph(gdf, type="distband", thresh=120)

        Fit a delaunay graph to a gdf without a dist threshold.
        >>> from cellseg_gsontools.graphs import fit_graph
        >>> w = fit_graph(gdf, type="delaunay", thresh=None)
    """
    allowed = (
        "delaunay",
        "knn",
        "distband",
        "relative_nhood",
    )
    if type not in allowed:
        raise ValueError(f"Illegal graph type given. Got: {type}. Allowed: {allowed}.")

    if gdf is None or gdf.empty:
        return

    # warn if id_col is not provided
    if not silence_warnings:
        _graph_warn(type, id_col)

    # can't fit delaunay or relative nhood graphs with less than 4 points
    if type in ("delaunay", "relative_nhood"):
        if len(gdf) < 4:
            return
    if type == "delaunay":
        # NOTE: neighbor keys start from 0
        w = Delaunay.from_dataframe(
            gdf.centroid,
            silence_warnings=True,
            use_index=True,
            ids=gdf[id_col],
            **kwargs,
        )
    elif type == "relative_nhood":
        # NOTE: neighbor indices start from 0
        w = Relative_Neighborhood.from_dataframe(
            gdf.centroid, silence_warnings=True, **kwargs
        )
    elif type == "knn":
        if "ids" in list(kwargs.keys()):  # drop ids kwarg since it fails
            kwargs.pop("ids")

        # NOTE: neighbor indices equal gdf[`ìd_col`]
        w = KNN.from_dataframe(gdf, silence_warnings=True, ids=id_col, **kwargs)
    elif type == "distband":
        if thresh is None:
            raise ValueError("DistBand requires `thresh` param. Not provided.")

        if "ids" in list(kwargs.keys()):  # drop ids kwarg since it fails
            kwargs.pop("ids")

        # NOTE: neighbor indices equal gdf[`ìd_col`]
        w = DistanceBand.from_dataframe(
            gdf,
            threshold=thresh,
            alpha=-1.0,
            ids=id_col,
            silence_warnings=True,
            **kwargs,
        )

    # convert graph indices to global ids
    if type in ("delaunay", "relative_nhood"):
        w = _graph_to_index(w, gdf)
        if id_col is not None:
            w = _graph_to_global_ids(w, gdf, id_col)

    # # Threshold edges based on distance to center node.
    if thresh is not None and type != "distband":
        w = dist_thresh_weights_sequential(gdf, w, thresh, id_col=id_col)

    return w


def dist_thresh_weights_sequential(
    gdf: gpd.GeoDataFrame,
    w: W,
    thresh: float,
    id_col: Optional[str] = None,
    include_self: bool = False,
) -> W:
    """Threshold edges based on distance to center node.

    Parameters:
        gdf (gpd.GeoDataFrame):
            The input geodataframe.
        w (libpysal.weights.W):
            The input spatial weights object.
        thresh (float):
            The distance threshold.
        id_col (str, optional):
            The unique id column in the gdf. If None, this uses `set_uid` to set it.
        include_self (bool, default=True):
            Whether to include self-loops in the neighbors.

    Returns:
        libpysal.weights.W:
            A libpysal spatial weights object, containing the neighbor graph data.
    """
    gdf = gdf.copy()

    # drop duplicate rows
    gdf = gdf.drop_duplicates(subset=[id_col], keep="first")
    gdf["nhood"] = pd.Series(list(w.neighbors.values()), index=gdf.index)

    new_neighbors = []
    for _, row in gdf.iterrows():
        neighbor_rows = gdf[gdf.loc[:, id_col].isin(row.nhood)]
        nhood_dist = [
            np.round(row.geometry.centroid.distance(ngh), 2)
            for ngh in neighbor_rows.geometry.centroid
        ]

        new_neighbors.append(
            _drop_neighbors(row.nhood, nhood_dist, thresh, include_self=include_self)
        )

    gdf["new_neighbors"] = new_neighbors
    return W(dict(zip(gdf[id_col], gdf["new_neighbors"])), silence_warnings=True)


def get_border_crosser_links(
    union_weights: W,
    roi_weights: W,
    iface_weights: W,
    only_border_crossers: bool = True,
) -> W:
    """Get the links that cross the border between the ROI and interface cells.

    Parameters:
        union_weights (W):
            The union of the ROI and interface weights. NOTE: contains links between ROI
            & interface cells.
        roi_weights (W):
            The ROI weights. NOTE: contains only links between ROI cells.
        iface_weights (W):
            The interface weights. NOTE: contains only links between interface cells.
        only_border_crossers (bool, optional):
            Whether to return only the links that cross the border between the ROI and
            interface cells or all neighbors of the node that has a border crossing link.
            This includes also the liks that do not cross the border. By default True.

    Returns:
        W:
            The links that cross the border between the ROI and interface cells.
    """
    # loop the nodes in the union graph and check if they are border crossing links
    graph = {}
    for node, neigh in union_weights.neighbors.items():
        is_roi_node = node in roi_weights.neighbors.keys()
        is_iface_node = node in iface_weights.neighbors.keys()

        neighbors = []
        for n in neigh:
            is_roi_neigh = n in roi_weights.neighbors.keys()
            is_iface_neigh = n in iface_weights.neighbors.keys()
            if (is_roi_node and is_iface_neigh) or (is_iface_node and is_roi_neigh):
                # return all neighbors if requested
                if not only_border_crossers:
                    neighbors.extend(neigh)
                    break
                else:
                    neighbors.append(n)

        graph[node] = neighbors

    return W(graph, silence_warnings=True)


def _drop_neighbors(
    nhood: Sequence[int],
    nhood_dist: Sequence[float],
    thresh: float,
    include_self: bool = False,
) -> np.ndarray:
    """Drop neighbors based on distance to center node."""
    if len(nhood) != len(nhood_dist):
        raise ValueError(
            f"Both inputs should be the same length. Got: `len(nhood) = {len(nhood)}`"
            f"and `len(nhood_dist) = {len(nhood_dist)}`."
        )

    nhood = np.array(nhood)
    nhood_dist = np.array(nhood_dist)

    # drop below thresh
    nhood = nhood[nhood_dist < thresh]

    # drop self loop
    if include_self:
        center = nhood[0]
        nhood = nhood[nhood != center]

    return nhood


def _graph_to_global_ids(w: W, in_gdf: gpd.GeoDataFrame, id_col: str) -> W:
    """Convert the graph ids to global ids i.e. to the ids of the id_col``."""
    new_neighbors = {}

    for node, neighbors in w.neighbors.items():
        new_id = in_gdf.loc[node, id_col]
        nghs = [in_gdf.loc[ngh, id_col] for ngh in neighbors]
        new_neighbors[new_id] = nghs

    return W(new_neighbors, id_order=sorted(new_neighbors.keys()))


def _graph_to_index(w: W, in_gdf: gpd.GeoDataFrame) -> W:
    new_neighbors = {}

    for node, neighbors in w.neighbors.items():
        new_id = in_gdf.iloc[node].name
        nghs = [in_gdf.iloc[ngh].name for ngh in neighbors]
        new_neighbors[new_id] = nghs

    return W(new_neighbors, id_order=sorted(new_neighbors.keys()))


def _graph_warn(type: str, id_col: str):
    if type in ("delaunay", "relative_nhood"):
        if id_col is None:
            warnings.warn(
                f"For graphs of type: {type}, if `id_col` is not provided "
                "The neighbors object will have keys starting from 0. "
                "These keys are likely not the same as the ids in the gdf."
            )
    elif type in ("knn", "distband"):
        if id_col is None:
            warnings.warn(
                f"For graphs of type: {type}, if `id_col` is not provided "
                "The neighbors object will have keys equalling the gdf.index. "
            )



from typing import Dict, List

import numpy as np
import scipy.ndimage as ndi
import skimage.filters as filters
from skimage import feature
from skimage.color import rgb2gray
from skimage.exposure import rescale_intensity
from skimage.feature import blob_dog
from skimage.util import invert

__all__ = ["intensity_props"]


def intensity_props(im: np.ndarray, mask: np.ndarray) -> Dict[str, np.ndarray]:
    """Computes intensity properties of an image mask.

    Parameters:
        im (np.ndarray):
            Image to compute properties from. Shape (H, W).
        mask (np.ndarray):
            Mask to apply to the image. Shape (H, W). Can be a label image.

    Returns:
        Dict[str, np.ndarray]:
            Dictionary with the following properties:
            - area
            - mean_intensity
            - std_intensity
            - max_intensity
            - min_intensity
            - median_intensity
    """
    intensity = rgb2gray(im)

    objs = ndi.find_objects(mask)
    res = {
        "area": np.zeros(len(objs)),
        "mean_intensity": np.zeros(len(objs)),
        "median_intensity": np.zeros(len(objs)),
        "std_intensity": np.zeros(len(objs)),
        "max_intensity": np.zeros(len(objs)),
        "min_intensity": np.zeros(len(objs)),
    }

    for i, sl in enumerate(objs):
        image = mask[sl] == i + 1
        _intensity = intensity[sl] * image
        res["area"][i] = np.sum(image)
        res["mean_intensity"][i] = np.mean(_intensity[image], axis=0)
        res["median_intensity"][i] = np.median(_intensity[image], axis=0)
        res["std_intensity"][i] = np.std(_intensity[image], axis=0)
        res["max_intensity"][i] = np.max(_intensity[image], axis=0)
        res["min_intensity"][i] = np.min(_intensity[image], axis=0)

    return res


def chromatin_feats(
    im: np.ndarray,
    mask: np.ndarray = None,
    invert_image: bool = False,
    features: List[str] = ["mask"],
) -> np.ndarray:
    """Computes chromatin features from a given nuclei image.

    Three features are available:
    - mask: Thresholded image.
    - blobs: Detected blobs.
    - edges: Detected edges.

    Thresholds an image based on chromatin content using the yen method.

    Parameters:
        im (np.ndarray):
            Image to threshold. Shape (H, W, 3).
        mask (np.ndarray):
            Mask to apply to the image. Shape (H, W).
        invert_image (bool):
            Invert the image before thresholding. This is useful for chromatin clumps.
            If not used, the lighter regions will be thresholded.
        features (List[str]):
            Allowed: "mask", "blobs", "edges"

    Returns:
        mask (np.ndarray):
             Thresholded image i.e. a binary mask. Shape (H, W).
        blobs (np.ndarray):
            Detected blobs. Shape (N, 3). First two columns are the coordinates.
            The third column is the radius.
        edges (np.ndarray):
            Detected edges i.e. a binary mask. Shape (H, W).
    """
    allowed_features = ["mask", "blobs", "edges"]
    if not all([f in allowed_features for f in features]):
        raise ValueError("Invalid feature. Allowed: 'mask', 'blobs', 'edges'")

    # invert to get chromatin clumps
    if invert_image:
        im = invert(im)

    # convert rgb to gray
    im = rescale_intensity(im)
    im = rgb2gray(im)

    # apply mask
    if mask is not None:
        im *= mask

    # apply threshold
    res = {}

    if "mask" in features:
        res["mask"] = im > filters.threshold_yen(im, nbins=2000)

    if "blobs" in features:
        blobs = blob_dog(im, max_sigma=30, threshold=0.1)
        blobs[:, 2] = blobs[:, 2] * np.sqrt(2)
        res["blobs"] = blobs

    if "edges" in features:
        res["edges"] = feature.canny(im, sigma=3)

    return res
